/*-
 * job.c --
 *	handle the creation etc. of our child processes.
 *
 * Copyright (c) 1988 by the Regents of the University of California
 * Copyright (c) 1988 by Adam de Boor
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appears in all copies.  Neither the University of California nor
 * Adam de Boor makes any representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 *
 *
 * Interface:
 *	Job_Make  	    	Start the creation of the given target.
 *
 *	Job_CatchChildren   	Check for and handle the termination of any
 *	    	  	    	children. This must be called reasonably
 *	    	  	    	frequently to keep the whole make going at
 *	    	  	    	a decent clip, since job table entries aren't
 *	    	  	    	removed until their process is caught this way.
 *	    	  	    	Its single argument is TRUE if the function
 *	    	  	    	should block waiting for a child to terminate.
 *
 *	Job_CatchOutput	    	Print any output our children have produced.
 *	    	  	    	Should also be called fairly frequently to
 *	    	  	    	keep the user informed of what's going on.
 *	    	  	    	If no output is waiting, it will block for
 *	    	  	    	a time given by the SEL_* constants, below,
 *	    	  	    	or until output is ready.
 *
 *	Job_Init  	    	Called to intialize this module. in addition,
 *	    	  	    	any commands attached to the .BEGIN target
 *	    	  	    	are executed before this function returns.
 *	    	  	    	Hence, the makefile must have been parsed
 *	    	  	    	before this function is called.
 *
 *	Job_Full  	    	Return TRUE if the job table is filled.
 *
 *	Job_Empty 	    	Return TRUE if the job table is completely
 *	    	  	    	empty.
 *
 *	Job_ParseShell	    	Given the line following a .SHELL target, parse
 *	    	  	    	the line as a shell specification. Returns
 *	    	  	    	FAILURE if the spec was incorrect.
 *
 *	Job_End	  	    	Perform any final processing which needs doing.
 *	    	  	    	This includes the execution of any commands
 *	    	  	    	which have been/were attached to the .END
 *	    	  	    	target. It should only be called when the
 *	    	  	    	job table is empty.
 *
 *	Job_AbortAll	    	Abort all currently running jobs. It doesn't
 *	    	  	    	handle output or do anything for the jobs,
 *	    	  	    	just kills them. It should only be called in
 *	    	  	    	an emergency, as it were.
 *
 *	Job_CheckCommands   	Verify that the commands for a target are
 *	    	  	    	ok. Provide them if necessary and possible.
 *
 *	Job_Touch 	    	Update a target without really updating it.
 *
 *	Job_Wait  	    	Wait for all currently-running jobs to finish.
 */
#ifndef lint
static char     *rcsid = "$Id: job.c,v 1.89 89/01/21 18:48:38 adam Exp $ SPRITE (Berkeley)";
#endif lint

#include    <stdio.h>
#include    <string.h>
#include    <sys/types.h>
#include    <sys/signal.h>
#include    <sys/stat.h>
#include    <fcntl.h>
#include    <sys/file.h>
#include    <sys/time.h>
#include    <sys/wait.h>
#include    <ctype.h>
#include    <errno.h>
extern int  errno;
#include    "make.h"

/*
 * Some systems define the fd_set we use, but not the macros to deal with it
 * (SunOS 3.5, e.g.)
 */
#ifndef FD_SET

# ifdef NEED_FD_SET
/*
 * Then there are the systems that don't even define fd_set...
 */
#  ifndef	FD_SETSIZE
#  define	FD_SETSIZE	256
#  endif
#  ifndef NBBY
#  define NBBY 8
#  endif

typedef long	fd_mask;
#define NFDBITS	(sizeof(fd_mask) * NBBY)	/* bits per mask */
#ifndef howmany
#define	howmany(x, y)	((unsigned int)(((x)+((y)-1)))/(unsigned int)(y))
#endif

typedef	struct fd_set {
	fd_mask	fds_bits[howmany(FD_SETSIZE, NFDBITS)];
} fd_set;

# endif /* NEED_FD_SET */

#define	FD_SET(n, p)	((p)->fds_bits[(n)/NFDBITS] |= (1 << ((n) % NFDBITS)))
#define	FD_CLR(n, p)	((p)->fds_bits[(n)/NFDBITS] &= ~(1 << ((n) % NFDBITS)))
#define	FD_ISSET(n, p)	((p)->fds_bits[(n)/NFDBITS] & (1 << ((n) % NFDBITS)))
#define FD_ZERO(p)	bzero((char *)(p), sizeof(*(p)))

#endif /* FD_SET */


#ifdef CUSTOMS
#include    "customs.h"
#endif /* CUSTOMS */

static void JobDoOutput();
static int JobStart();
static void JobRestart();
static void JobFinish();
static int JobInterrupt ();

/*
 * The SEL_ constants determine the maximum amount of time spent in select
 * before coming out to see if a child has finished. SEL_SEC is the number of
 * seconds and SEL_USEC is the number of micro-seconds 
 */
#define SEL_SEC		0
#define SEL_USEC	500000

/*
 * error handling variables 
 */
static int        errors = 0;	    /* number of errors reported */
static int  	  aborting = 0;	    /* why is the make aborting? */
#define ABORT_ERROR	1   	    /* Because of an error */
#define ABORT_INTERRUPT	2   	    /* Because it was interrupted */
#define ABORT_WAIT	3   	    /* Waiting for jobs to finish */


/*
 * post-make command processing. The node postCommands is really just the
 * .END target but we keep it around to avoid having to search for it
 * all the time.
 */
static GNode   	  *postCommands;    /* node containing commands to execute when
				     * everything else is done */
static int     	  numCommands; 	    /* The number of commands actually printed
				     * for a target. Should this number be
				     * 0, no shell will be executed. */


/*
 * Return values from JobStart.
 */
#define JOB_RUNNING	0   	/* Job is running */
#define JOB_ERROR 	1   	/* Error in starting the job */
#define JOB_FINISHED	2   	/* The job is already finished */
#define JOB_STOPPED	3   	/* The job is stopped */

/*
 * tfile is the name of a file into which all shell commands are put. It is
 * used over by removing it before the child shell is executed. The XXXXX in
 * the string are replaced by the pid of the make process in a 5-character
 * field with leading zeroes. 
 */
static char     tfile[] = "/tmp/makeXXXXX";


/*-
 * Shell Specifications:
 * Each shell type has associated with it the following information:
 *	1) The string which must match the last character of the shell name
 *	   for the shell to be considered of this type. The longest match
 *	   wins.
 *	2) A command to issue to turn off echoing of command lines
 *	3) A command to issue to turn echoing back on again
 *	4) What the shell prints, and its length, when given the echo-off
 *	   command. This line will not be printed when received from the shell
 *	5) A boolean to tell if the shell has the ability to control
 *	   error checking for individual commands.
 *	6) The string to turn this checking on.
 *	7) The string to turn it off.
 *	8) The command-flag to give to cause the shell to start echoing
 *	   commands right away.
 *	9) The command-flag to cause the shell to Lib_Exit when an error is
 *	   detected in one of the commands.
 *
 * Some special stuff goes on if a shell doesn't have error control. In such
 * a case, errCheck becomes a printf template for echoing the command,
 * should echoing be on and ignErr becomes another printf template for
 * executing the command while ignoring the return status. If either of these
 * strings is empty when hasErrCtl is FALSE, the command will be executed
 * anyway as is and if it causes an error, so be it.
 */
typedef struct Shell {
    char	  *name;	/* the name of the shell. For Bourne and C
				 * shells, this is used only to find the
				 * shell description when used as the single
				 * source of a .SHELL target. For user-defined
				 * shells, this is the full path of the shell.
				 */
    Boolean 	  hasEchoCtl;	/* True if both echoOff and echoOn defined */
    char          *echoOff;	/* command to turn off echo */
    char          *echoOn;	/* command to turn it back on again */
    char          *noPrint;	/* command to skip when printing output from
				 * shell. This is usually the command which
				 * was executed to turn off echoing */
    int           noPLen;	/* length of noPrint command */
    Boolean	  hasErrCtl;	/* set if can control error checking for
				 * individual commands */
    char	  *errCheck;	/* string to turn error checking on */
    char	  *ignErr;	/* string to turn off error checking */
    /*
     * command-line flags 
     */
    char          *echo;	/* echo commands */
    char          *exit;	/* exit on error */
}               Shell;

static Shell    shells[] = {
    /*
     * CSH description. The csh can do echo control by playing
     * with the setting of the 'echo' shell variable. Sadly,
     * however, it is unable to do error control nicely.
     */
{
    "csh",
    TRUE, "unset verbose", "set verbose", "unset verbose", 10,
    FALSE, "echo \"%s\"\n", "csh -c \"%s || exit 0\"",
    "v", "e",
},
    /*
     * SH description. Echo control is also possible and, under
     * sun UNIX anyway, one can even control error checking.
     */
{
    "sh",
    TRUE, "set -", "set -v", "set -", 5,
#if (defined(sun) && !defined(Sprite)) || defined(SYSV)
    TRUE, "set -e", "set +e",
#else
    FALSE, "echo \"%s\"\n", "sh -c '%s || exit 0'\n",
#endif
    "v", "e",
},
    /*
     * UNKNOWN.
     */
{
    (char *)0,
    FALSE, (char *)0, (char *)0, (char *)0, 0,
    FALSE, (char *)0, (char *)0,
    (char *)0, (char *)0,
}
};
static Shell   *commandShell = &shells[DEFSHELL]; /* this is the shell to
						   * which we pass all
						   * commands in the Makefile.
						   * It is set by the
						   * Job_ParseShell function */
static char    *shellPath = (char *) NULL,	  /* full pathname of
						   * executable image */
               *shellName;	      	      	  /* last component of shell */


/*-
 * Job Table definitions. 
 *
 * Each job has several things associated with it:
 *	1) The process id of the child shell
 *	2) The graph node describing the target being made by this job
 *	3) A LstNode for the first command to be saved after the job
 *	   completes. This is NILLNODE if there was no "..." in the job's
 *	   commands.
 *	4) An FILE* for writing out the commands. This is only
 *	   used before the job is actually started.
 *	5) A union of things used for handling the shell's output. Different
 *	   parts of the union are used based on the value of the usePipes
 *	   flag. If it is true, the output is being caught via a pipe and
 *	   the descriptors of our pipe, an array in which output is line
 *	   buffered and the current position in that buffer are all
 *	   maintained for each job. If, on the other hand, usePipes is false,
 *	   the output is routed to a temporary file and all that is kept
 *	   is the name of the file and the descriptor open to the file.
 *	6) An identifier provided by and for the exclusive use of the
 *	   Rmt module.
 *	7) A word of flags which determine how the module handles errors,
 *	   echoing, etc. for the job 
 *
 * The job "table" is kept as a linked Lst in 'jobs', with the number of
 * active jobs maintained in the 'nJobs' variable. At no time will this
 * exceed the value of 'maxJobs', initialized by the Job_Init function. 
 *
 * When a job is finished, the Make_Update function is called on each of the
 * parents of the node which was just remade. This takes care of the upward
 * traversal of the dependency graph.
 */
#define JOB_BUFSIZE	1024
typedef struct Job {
    int       	  pid;	    /* The child's process ID */
    GNode    	  *node;    /* The target the child is making */
    LstNode 	  tailCmds; /* The node of the first command to be
			     * saved when the job has been run */
    FILE 	  *cmdFILE; /* When creating the shell script, this is
			     * where the commands go */
#ifdef CUSTOMS
    ExportPermit  permit;   	/* Permit under which job was exported */
    int	    	  rmtFd;    	/* FD of stream to remote job. Needed if
				 * usePipes is FALSE */
    int	    	  cmdFd;  	/* FD of file containing commands */
#endif /* CUSTOMS */
    int	    	  rmtID;    	/* ID returned from Rmt module */
    short      	  flags;	/* Flags to control treatment of job */
#define	JOB_IGNERR	0x001	/* Ignore non-zero exits */
#define	JOB_SILENT	0x002	/* no output */
#define JOB_SPECIAL	0x004	/* Target is a special one. i.e. run it locally
				 * if we can't export it and maxLocal is 0 */
#define JOB_IGNDOTS	0x008  	/* Ignore "..." lines when processing
				 * commands */
#define JOB_REMOTE	0x010	/* Job is running remotely */  
#define JOB_FIRST	0x020	/* Job is first job for the node */
#define JOB_REMIGRATE	0x040	/* Job needs to be remigrated */
#define JOB_RESTART	0x080	/* Job needs to be completely restarted */
#define JOB_RESUME	0x100	/* Job needs to be resumed b/c it stopped,
				 * for some reason */
#define JOB_CONTINUING	0x200	/* We are in the process of resuming this job.
				 * Used to avoid infinite recursion between
				 * JobFinish and JobRestart */
    union {
	struct {
	    int	  	    op_inPipe;	/* Input side of pipe associated
					 * with job's output channel */
	    int   	    op_outPipe;	/* Output side of pipe associated with
					 * job's output channel */
	    char  	    op_outBuf[JOB_BUFSIZE + 1];
	    	  	    	    	/* Buffer for storing the output of the
					 * job, line by line */
	    int   	    op_curPos;	/* Current position in op_outBuf */
	}         	o_pipe;	    /* data used when catching the output via
				     * a pipe */
	struct {
	    char  	    of_outFile[sizeof(tfile)+2];
	    	  	    	    	/* Name of file to which shell output
					 * was rerouted */
	    int	  	    of_outFd;	/* Stream open to the output
					 * file. Used to funnel all
					 * from a single job to one file
					 * while still allowing
					 * multiple shell invocations */
	}   	  	o_file;	    /* Data used when catching the output in
				     * a temporary file */
    }          	  output;	/* Data for tracking a shell's output */

} Job;

#define outPipe	  	output.o_pipe.op_outPipe
#define inPipe	  	output.o_pipe.op_inPipe
#define outBuf		output.o_pipe.op_outBuf
#define curPos		output.o_pipe.op_curPos
#define outFile		output.o_file.of_outFile
#define outFd	  	output.o_file.of_outFd

static int  	  maxJobs;    	/* The most children we can run at once */
static int  	  maxLocal;    	/* The most local ones we can have */
static int  	  nJobs;	/* The number of children currently running */
static int  	  nLocal;    	/* The number of local children */
static Lst  	  jobs;		/* The structures which describe them */
static Boolean	  jobFull;    	/* Flag to tell when the job table is full. It
				 * is set TRUE when (1) the total number of
				 * running jobs equals the maximum allowed or
				 * (2) a job can only be run locally, but
				 * nLocal equals maxLocal */
static fd_set  	  outputs;    	/* Set of descriptors of pipes connected to
				 * the output channels of children */
static GNode 	  *lastNode;	/* The node for which output was most recently
				 * produced. */
#define TARG_FMT  "--- %s ---\n"

/*
 * When JobStart attempts to run a job remotely but can't, and isn't allowed
 * to run the job locally, or when Job_CatchChildren detects a job that has
 * been migrated home, the job is placed on the stoppedJobs queue to be run
 * when the next job finishes. 
 */
static Lst  	  stoppedJobs;	/* Lst of Job structures describing
				 * jobs that were stopped due to concurrency
				 * limits or migration home */


#ifdef CUSTOMS
/*
 * This stuff is all for using Customs remote execution. It could be put in its
 * own rmt.c if the Job structure were exported and a new interface defined for
 * waiting on inputs and so on, but it seems more trouble than it's worth.
 */


static char 	  	cwd[1024];  /* The current working directory */
static Boolean	  	noAgent;    /* TRUE if agent not present */

/*
 * From customslib.c
 */
extern int  	  	customs_Socket;	/* Socket opened by Customs calls */
extern struct timeval	customs_RetryTimeOut;	/* Default retry interval */
extern struct sockaddr_in customs_AgentAddr;

/*
 * For Make-mode exportation.
 */
static int returnFD;
static int exportFD;
static ExportPermit permit;

#include    <sys/ioctl.h>
#include    <netdb.h>

/*-
 *-----------------------------------------------------------------------
 * JobSwapExit --
 *	Byte-swap an Exit_Data structure.
 *
 * Results:
 *	None.
 *	
 * Side Effects:
 *	The words in the Exit_Data structure are byte-swapped.
 *
 *-----------------------------------------------------------------------
 */
static void
JobSwapExit(length, data)
    int	    	  length;
    Exit_Data	  *data;
{
    Rpc_SwapLong(sizeof(long), &data->id);
    Rpc_SwapLong(sizeof(long), &data->status);
}

/*-
 *-----------------------------------------------------------------------
 * JobLocalInput --
 *	Handle a pipe becoming readable. Callback function for Rpc_Watch
 *
 * Results:
 *	None
 *
 * Side Effects:
 *	JobDoOutput is called.
 *	
 *-----------------------------------------------------------------------
 */
/*ARGSUSED*/
static void
JobLocalInput(stream, job, state)
    int	    stream;
    Job	    *job;
    int	    state;
{
    if (state & RPC_READABLE) {
	JobDoOutput(job, FALSE);
    }
}

static Boolean printEm = FALSE;	/* Set true if debugging and JobRmtCmpID
				 * should print out the Job structures as it
				 * checks them. */


/*-
 *-----------------------------------------------------------------------
 * JobRmtCmpID --
 *	See if a job is remote and has the given ID number
 *
 * Results:
 *	0 if it matches, non-zero if it doesn't
 *
 * Side Effects:
 *	None
 *-----------------------------------------------------------------------
 */
static int
JobRmtCmpID(job, id)
    Job	    	  *job;
    unsigned long id;
{
    if (printEm) {
	printf("\t%s: ", job->node->name);
	if (job->flags & JOB_REMOTE) {
	    printf("remote #%d\n", job->permit.id);
	} else {
	    printf("local #%d\n", job->pid);
	}
    }
    if (job->flags & JOB_REMOTE) {
	return (id - job->permit.id);
    } else {
	return(1);
    }
}
/*-
 *-----------------------------------------------------------------------
 * JobRmtExit --
 *	Handle a call on the returnFD to tell us a job has exited.
 *
 * Results:
 *	Nothing.
 *
 * Side Effects:
 *	A Job is finished out.
 *	
 *-----------------------------------------------------------------------
 */
static void
JobRmtExit(from, msg, len, data)
    struct sockaddr_in	*from;	    /* Call from... */
    Rpc_Message	  	msg;	    /* Token for return */
    int			len;	    /* Length of passed data */
    Rpc_Opaque		data;	    /* Data passed */
{
    Exit_Data	  	*eData;	    /* Data in our format */
    register LstNode	ln; 	    /* Node of finished job */
    register Job  	*job;	    /* The job itself */
    union wait	  	status;	    /* How it died */

    if (msg) {
	/*
	 * Acknowledge the call
	 */
	Rpc_Return(msg, 0, (Rpc_Opaque)0);
    }

    eData = (Exit_Data *)data;
    status.w_status = eData->status;

    if (WIFSTOPPED(status)) {
	/*
	 * If the remote job has stopped, it must be because of us -- we've
	 * already continued it, therefore (or will, anyway), so there's no
	 * reason to slap the thing onto the stopped list and confuse things
	 * terribly...so we don't.
	 */
	return;
    }

    ln = Lst_Find(jobs, eData->id, JobRmtCmpID);
    if (ln == NILLNODE) {
	ln = Lst_Find(stoppedJobs, eData->id, JobRmtCmpID);
	if (ln != NILLNODE) {
	    Error("Received exit for stopped job, id %d\n", eData->id);
	} else {
	    Error("Received exit for unknown job id %d from %s\n", eData->id,
		  inet_ntoa(from->sin_addr));
	    printEm = TRUE;
	    (void)Lst_Find(jobs, eData->id, JobRmtCmpID);
	    printEm = FALSE;
	}
	return;
    }
    job = (Job *)Lst_Datum(ln);
    if (!usePipes) {
	/*
	 * Flush all data from the socket into the file before calling
	 * JobFinish
	 */
	int	  	nb;
	char  	buf[512];
	
	while ((ioctl(job->rmtFd, FIONREAD, &nb) == 0) && (nb > 0)) {
	    if (nb > sizeof(buf)) {
		nb = sizeof(buf);
	    }
	    nb = read(job->rmtFd, buf, nb);
	    write(job->outFd, buf, nb);
	}
	close(job->rmtFd);
    }
    
    /*
     * Take it out of the list
     */
    (void)Lst_Remove(jobs, ln);
    nJobs -= 1;
    
    /*
     * Table no longer full
     */
    jobFull = FALSE;
    
    /*
     * Finish it out
     */
    JobFinish(job, eData->status);
    
    /*
     * Set aborting if any error.
     */
    if (errors && !keepgoing && (aborting != ABORT_INTERRUPT)) {
	/*
	 * If we found any errors in this batch of children and the -k flag
	 * wasn't given, we set the aborting flag so no more jobs get
	 * started.
	 */
	aborting = ABORT_ERROR;
    }
    
    if ((aborting == ABORT_ERROR) && Job_Empty()) {
	/*
	 * If we are aborting and the job table is now empty, we finish.
	 */
	(void) unlink (tfile);
	Finish (errors);
    }
}

/*-
 *-----------------------------------------------------------------------
 * JobRmtIO  --
 *	Handle I/O transfers between a remote job and the local machine.
 *
 * Results:
 *	Nothing.
 *
 * Side Effects:
 *	None
 *-----------------------------------------------------------------------
 */
static Boolean JobBlockTimeout() { return (TRUE); }
/*ARGSUSED*/
static void
JobRmtIO(stream, job, state)
    int	    	  stream;
    Job		  *job;
    int		  state;
{
    if (state & RPC_READABLE) {
	if (usePipes) {
	    /*
	     * If we're using pipes, we can just use JobDoOutput to transfer
	     * the data to the screen.
	     */
	    JobDoOutput(job, FALSE);
	} else {
	    /*
	     * Otherwise, we have to actually write the data to the output
	     * file. Just read a single 1K block each time through.
	     * Magic numbers R us.
	     */
	    char  	buf[1024];
	    int		nb;

	    nb = read(job->rmtFd, buf, sizeof(buf));
	    if (nb > 0) {
		write(job->outFd, buf, nb);
	    }
	}
    }
    if (state & RPC_WRITABLE) {
	char	  buf[512];
	int 	  nb;

	nb = read(job->cmdFd, buf, sizeof(buf));
	if (nb > 0) {
	    char  *cp;
	    struct timeval tv;
	    Rpc_Event ev;

	    /*
	     * Stop watching for this stream to be writable until this buffer
	     * is written out
	     */
	    Rpc_Watch(job->rmtFd, RPC_READABLE, JobRmtIO, (ClientData)job);

	    /*
	     * Write out the buffer in whatever chunks the socket can swallow
	     */
	    cp = buf;

	    /*
	     * Create a timeout event to be invoked every 200 ms that makes
	     * sure Rpc_Wait will return with reasonable speed.
	     */
	    tv.tv_sec = 0;
	    tv.tv_usec = 200000;
	    ev = Rpc_EventCreate(&tv, JobBlockTimeout, (Rpc_Opaque)0);
	    
	    while (nb > 0) {
		int	cc = write(job->rmtFd, cp, nb);

		if (cc < 0) {
		    break;
		} else {
		    cp += cc;
		    nb -= cc;
		    if (nb > 0) {
			/*
			 * Give other things a chance while we wait for the
			 * socket to drain enough.
			 */
			Rpc_Wait();
		    }
		}
	    }
	    Rpc_EventDelete(ev);
	    Rpc_Watch(job->rmtFd, RPC_READABLE|RPC_WRITABLE, JobRmtIO,
		      (ClientData)job);
	    
	    if (nb > 0) {
		Exit_Data   eData;
		
		if (lastNode != job->node) {
		    printf(TARG_FMT, job->node->name);
		    lastNode = job->node;
		}
		if (errno == EPIPE) {
		    printf("*** connection closed\n");
		} else {
		    perror("*** writing to remote");
		}
		eData.id = job->permit.id;
		eData.status = 1;
		JobRmtExit((struct sockaddr_in *)0, (Rpc_Message)0,
			   sizeof(eData), (Rpc_Opaque)&eData);
	    }
	} else {
	    if (nb < 0) {
		perror("*** read(cmd)");
	    }
	    /*
	     * Nothing more to read, so force an EOF on the other side
	     * by doing an out-going shutdown, then only pay attention
	     * to the beast being readable.
	     */
	    shutdown(job->rmtFd, 1);
	    Rpc_Watch(job->rmtFd, RPC_READABLE, JobRmtIO, (Rpc_Opaque)job);
	}
    }
}

/*-
 *-----------------------------------------------------------------------
 * JobCheckChildren --
 *	Timeout handler for seeing if anyone has died. Simply calls
 *	Job_CatchChildren with block set to FALSE, then returns TRUE to
 *	force Rpc_Wait to exit.
 *
 * Results:
 *	TRUE.
 *
 * Side Effects:
 *	Job_CatchChildren is called.
 *	
 *-----------------------------------------------------------------------
 */
static Boolean
JobCheckChildren()
{
    int	    oldnj = nJobs;
    
    Job_CatchChildren(FALSE);

    return(oldnj != nJobs);
}

/*-
 *-----------------------------------------------------------------------
 * PassSigIfRemote --
 *	Pass a signal to a job only if the job is remote.
 *
 * Results:
 *	=== 0
 *
 * Side Effects:
 *	An Rpc is issued to the server for the job to kill the thing
 *	with the same signal.
 *	
 *-----------------------------------------------------------------------
 */
static int
PassSigIfRemote(job, signo)
    Job	    	  *job;
    int	    	  signo;
{
    if (job->flags & JOB_REMOTE) {
	Kill_Data 	    packet;
	struct sockaddr_in  server;
	Rpc_Stat  	    rstat;

	server.sin_family = AF_INET;
	server.sin_addr = job->permit.addr;
	server.sin_port = customs_AgentAddr.sin_port;
	packet.id = job->permit.id;
	packet.signo = signo;

	rstat = Rpc_Call(customs_Socket, &server, (Rpc_Proc)CUSTOMS_KILL,
			 sizeof(packet), (Rpc_Opaque)&packet,
			 0, (Rpc_Opaque)0,
			 CUSTOMS_NRETRY, &customs_RetryTimeOut);
	if (rstat != RPC_SUCCESS) {
	    printf("%s: %s\n", job->node->name, Rpc_ErrorMessage(rstat));
	}
    }
    return(0);
}

/*-
 *-----------------------------------------------------------------------
 * PassSig --
 *	Pass a signal on to all remote jobs.
 *
 * Results:
 *	None.
 *
 * Side Effects:
 *	We die by the same signal.
 *	
 *-----------------------------------------------------------------------
 */
static void
PassSig(signo)
    int	    signo;	/* The signal number we've received */
{
    int	    mask;
    
    Lst_ForEach(jobs, PassSigIfRemote, (ClientData)signo);

    mask = sigblock(0);
    (void) sigsetmask(mask & ~(1 << (signo-1)));
    signal(signo, SIG_DFL);

    kill(getpid(), signo);

    sigsetmask(mask);
    signal(signo, PassSig);

    Lst_ForEach(jobs, PassSigIfRemote, (ClientData)SIGCONT);
}

/*-
 *-----------------------------------------------------------------------
 * Rmt_Init --
 *	Initialize this module...
 *
 * Results:
 *	None.
 *
 * Side Effects:
 *	The current working directory is placed in cwd and noAgent set
 *	FALSE if the local customs agent could be contacted.
 *
 *-----------------------------------------------------------------------
 */
void
Rmt_Init()
{
    Rpc_Stat	  status;

    if (noExport || ((status = Customs_Ping()) != RPC_SUCCESS)) {
	if (DEBUG(RMT) && !noExport) {
	    printf("Could not contact customs agent: %s\n",
		   Rpc_ErrorMessage(status));
	}
	noAgent = TRUE;
    } else {
	noAgent = FALSE;
	    
	Rpc_ServerCreate(customs_Socket, (Rpc_Proc)CUSTOMS_EXIT,
			 JobRmtExit, JobSwapExit, Rpc_SwapNull,
			 (Rpc_Opaque)0);
	    
	getwd(cwd);
	if (DEBUG(RMT)) {
	    printf("Customs agent present. cwd = \"%s\"\n", cwd);
	}
	if (!amMake) {
	    signal (SIGHUP, PassSig);
	    signal (SIGQUIT, PassSig);
	    signal (SIGTERM, PassSig);
	    signal (SIGTSTP, PassSig);
	    signal (SIGTTOU, PassSig);
	    signal (SIGTTIN, PassSig);
	    signal (SIGWINCH, PassSig);
		
	    signal (SIGPIPE, SIG_IGN); /* This isn't right */
	}
    }
}

/*-
 *-----------------------------------------------------------------------
 * Rmt_AddServer --
 *	Add a server to the list of those known.
 *
 * Results:
 *	None.
 *
 * Side Effects:
 *	None.
 *
 *-----------------------------------------------------------------------
 */
/*ARGSUSED*/
void
Rmt_AddServer (name)
    char    *name;
{
}
/*-
 *-----------------------------------------------------------------------
 * Rmt_Begin --
 *	Prepare to export a job -- the Make-mode interface to Customs.
 *
 * Results:
 *	TRUE if the job can be exported. FALSE if it cannot.
 *
 * Side Effects:
 *	A TCP connection is opened to an available server and the
 *	CUSTOMS_IMPORT command issued (i.e. the job is started
 *	over there). exportFD is set to the fd of the connection and
 *	returnFD to the fd of the socket to be used to return the exit
 *	status.
 *
 *-----------------------------------------------------------------------
 */
/*ARGSUSED*/
Boolean
Rmt_Begin (file, argv, gn)
    char    	  *file;
    char    	  **argv;
    GNode   	  *gn;
{
    if (noAgent) {
	return (FALSE);
    } else {
	extern int errno;
	returnFD = -1;
	errno = 0;

	if (DEBUG(RMT)) {
	    printf("Rmt_Begin: sameArch = %s\n",
		   (gn->type & OP_EXPORTSAME) ? "True" : "False");
	}
	exportFD = Customs_RawExport(file, argv, cwd,
				     ((gn->type & OP_EXPORTSAME) ?
				      EXPORT_SAME :
				      EXPORT_ANY),
				     &returnFD,
				     &permit);
	if (exportFD < 0) {
	    if (DEBUG(RMT)) {
		perror("Customs_RawExport");
		Customs_PError(file);
	    }
	    return (FALSE);
	} else {
	    struct hostent *he;

	    he = gethostbyaddr(&permit.addr,
			       sizeof(permit.addr),
			       AF_INET);
	    if (he == (struct hostent *)NULL) {
		printf("*** exported to %s (id %u)\n",
		       inet_ntoa(permit.addr),
		       permit.id);
	    } else {
		printf("*** exported to %s (id %u)\n", he->h_name,
		       permit.id);
	    }
	    
	    return (TRUE);
	}
    }
}

/*-
 *-----------------------------------------------------------------------
 * Rmt_Exec --
 *	Execute a process elsewhere. If the exportation actually succeeded
 *	(exportFD > 0), the "export" program is executed (must be on the
 *	search path) with the -id flag.
 *
 * Results:
 *	None.
 *
 * Side Effects:
 *	That remains to be seen.
 *
 *-----------------------------------------------------------------------
 */
void
Rmt_Exec (file, args, traceMe)
    char    *file;
    char    **args;
    Boolean traceMe;
{
    if (exportFD > 0) {
	char	  fd1[4], fd2[4];
	char	  id[10];
	char	  *argv[6];

	sprintf (fd1, "%d", exportFD);
	sprintf (fd2, "%d", returnFD);
	sprintf (id, "%08x", permit.id);

	argv[0] = "export to";
	argv[1] = "-id";
	argv[2] = fd1;
	argv[3] = fd2;
	argv[4] = id;
	argv[5] = (char *)0;
	if (DEBUG(RMT)) {
	    printf("export -id %s %s %s\n", argv[2], argv[3], argv[4]);
	}
	(void)execvp ("export", argv);
	if (DEBUG(RMT)) {
	    perror("Couldn't exec \"export\"");
	}
    } else {
	if (DEBUG(RMT)) {
	    printf("Rmt_Exec called when exportFD == %d\n", exportFD);
	}
	(void)execvp (file, args);
    }
}

/*-
 *-----------------------------------------------------------------------
 * JobExport --
 *	Prepare to export a job -- the PMake-mode interface to customs.
 *
 * Results:
 *	TRUE if the job can be exported. FALSE if it cannot.
 *
 * Side Effects:
 *	A TCP connection is opened to an available server and the
 *	CUSTOMS_IMPORT command issued (i.e. the job is started
 *	over there). job->rmtFd is set to the fd of the connection and
 *	job->permit to the permit under which the job is running. If
 *	usePipes is TRUE, the pipes that were opened in JobStart are
 *	closed again and job->inPipe is set to job->rmtFd to allow
 *	us to use JobDoOutput to handle the output.
 *
 *-----------------------------------------------------------------------
 */
/*ARGSUSED*/
static Boolean
JobExport (file, argv, job)
    char    	  *file;    	    	/* File to exec */
    char    	  **argv;   	    	/* Arguments to it */
    Job   	  *job;	    	    	/* Job descriptor for it */
{
    if (noAgent) {
	return (FALSE);
    } else {
	extern int errno;
	struct timeval now, then;

	errno = 0;

	if (DEBUG(RMT)) {
	    printf("JobExport: sameArch = %s\n",
		   (job->node->type & OP_EXPORTSAME) ? "True" : "False");
	    gettimeofday(&now, 0);
	}

	job->rmtFd = Customs_RawExport(file, argv, cwd,
				       ((job->node->type & OP_EXPORTSAME) ?
					EXPORT_SAME :
					EXPORT_ANY),
				       &customs_Socket,
				       &job->permit);
	if (DEBUG(RMT)) {
	    gettimeofday(&then, 0);
	    then.tv_usec -= now.tv_usec;
	    if (then.tv_usec < 0) {
		then.tv_usec += 1000000;
		then.tv_sec -= now.tv_sec + 1;
	    } else {
		then.tv_sec -= now.tv_sec;
	    }
	    printf("*** time spent calling: %d.%06d s\n",
		   then.tv_sec, then.tv_usec);
	}

	if (job->rmtFd < 0) {
	    if (DEBUG(RMT)) {
		perror("Customs_RawExport");
		Customs_PError(file);
	    } else {
		if (job->rmtFd > -100) {
		    if (usePipes) {
			printf (TARG_FMT, job->node->name);
			lastNode = job->node;
		    }
		    printf("*** error calling local server: %s\n",
			   Rpc_ErrorMessage(-job->rmtFd));
		} else switch(job->rmtFd) {
		case CUSTOMS_NOEXPORT:
		    /*
		     * Just couldn't export -- no biggie.
		     */
		    break;
		case CUSTOMS_NORETURN:
		    /*
		     * Couldn't create return socket
		     */
		    if (usePipes) {
			printf (TARG_FMT, job->node->name);
			lastNode = job->node;
		    }
		    printf("*** error exporting: couldn't create return socket\n");
		    break;
		case CUSTOMS_NONAME:
		    /*
		     * Couldn't fetch name of socket
		     */
		    if (usePipes) {
			printf (TARG_FMT, job->node->name);
			lastNode = job->node;
		    }
		    printf("*** error exporting: couldn't find name of return socket\n");
		    break;
		case CUSTOMS_ERROR:
		    if (usePipes) {
			printf (TARG_FMT, job->node->name);
			lastNode = job->node;
		    }
		    printf("*** error exporting\n");
		    break;
		default:
		    if (usePipes) {
			printf (TARG_FMT, job->node->name);
			lastNode = job->node;
		    }
		    printf("*** error calling remote server: %s\n",
			   Rpc_ErrorMessage(-(job->rmtFd + 200)));
		    break;
		}
		fflush(stdout);
	    }
	    return (FALSE);
	} else {
	    struct hostent *he;

	    he = gethostbyaddr(&job->permit.addr,
			       sizeof(job->permit.addr),
			       AF_INET);
	    if (usePipes) {
		printf (TARG_FMT, job->node->name);
		lastNode = job->node;
	    }
	    
	    if (he == (struct hostent *)NULL) {
		printf("*** exported to %s (id %u)\n",
		       inet_ntoa(job->permit.addr),
		       job->permit.id);
	    } else {
		printf("*** exported to %s (id %u)\n", he->h_name,
		       job->permit.id);
	    }
	    
	    if (usePipes) {
		/*
		 * Close down the pipes that were opened for this job since
		 * we dinnae need them
		 */
		close(job->outPipe);
		close(job->inPipe);
		job->outPipe = job->inPipe = job->rmtFd;
		if (DEBUG(RMT)) {
		    printf("rmtFd = %d\n", job->rmtFd);
		}
	    }
	    
	    fflush(stdout);

	    fcntl(job->rmtFd, F_SETFL, FNDELAY);
	    fcntl(job->rmtFd, F_SETFD, 1);

	    return (TRUE);
	}
    }
}

/*-
 *-----------------------------------------------------------------------
 * Rmt_ReExport --
 *	Supposed to re-export a job that's come home, but since jobs
 *	can't come home under customs, we just return FALSE to say
 *	we couldn't do it.
 *
 * Results:
 *	FALSE.
 *
 * Side Effects:
 *	None.
 *
 *-----------------------------------------------------------------------
 */
/*ARGSUSED*/
Boolean
Rmt_ReExport(pid)
    int	    pid;
{
    if (DEBUG(RMT)) {
	printf("Rmt_ReExport called?\n");
    }
    return(FALSE);
}

/*-
 *-----------------------------------------------------------------------
 * Rmt_LastID --
 *	Return an unique identifier for the last job exported with Rmt_Exec
 *
 * Results:
 *	Some sort of identifier. Just returns 1.
 *
 * Side Effects:
 *	returnFD and exportFD are closed if we're in Make mode.
 *
 *-----------------------------------------------------------------------
 */
/*ARGSUSED*/
int
Rmt_LastID(pid)
    int	    	  pid;	    /* PID of job last exported */
{
    if (amMake) {
	(void)close(returnFD);
	(void)close(exportFD);
	exportFD = 0;
    }
    return (1);
}

/*-
 *-----------------------------------------------------------------------
 * Rmt_Done --
 *	Register the completion of a remote job.
 *
 * Results:
 *	None.
 *
 * Side Effects:
 *	None.
 *
 *-----------------------------------------------------------------------
 */
/*ARGSUSED*/
void
Rmt_Done (id)
    int	    id;
{
}

#endif /* CUSTOMS */


/*-
 *-----------------------------------------------------------------------
 * JobCmpPid  --
 *	Compare the pid of the job with the given pid and return 0 if they
 *	are equal. This function is called from Job_CatchChildren via
 *	Lst_Find to find the job descriptor of the finished job.
 *
 * Results:
 *	0 if the pid's match
 *
 * Side Effects:
 *	None
 *-----------------------------------------------------------------------
 */
static int
JobCmpPid (job, pid)
    int             pid;	/* process id desired */
    Job            *job;	/* job to examine */
{
    return (pid - job->pid);
}

/*-
 *-----------------------------------------------------------------------
 * JobPrintCommand  --
 *	Put out another command for the given job. If the command starts
 *	with an @ or a - we process it specially. In the former case,
 *	so long as the -s and -n flags weren't given to make, we stick
 *	a shell-specific echoOff command in the script. In the latter,
 *	we ignore errors for the entire job, unless the shell has error
 *	control.
 *	If the command is just "..." we take all future commands for this
 *	job to be commands to be executed once the entire graph has been
 *	made and return non-zero to signal that the end of the commands
 *	was reached. These commands are later attached to the postCommands
 *	node and executed by Job_End when all things are done.
 *	This function is called from JobStart via Lst_ForEach.
 *
 * Results:
 *	Always 0, unless the command was "..."
 *
 * Side Effects:
 *	If the command begins with a '-' and the shell has no error control,
 *	the JOB_IGNERR flag is set in the job descriptor.
 *	If the command is "..." and we're not ignoring such things,
 *	tailCmds is set to the successor node of the cmd.
 *	numCommands is incremented if the command is actually printed.
 *-----------------------------------------------------------------------
 */
static int
JobPrintCommand (cmd, job)
    char     	  *cmd;	    	    /* command string to print */
    Job           *job;	    	    /* job for which to print it */
{
    Boolean	  noSpecials;	    /* true if we shouldn't worry about
				     * inserting special commands into
				     * the input stream. */
    Boolean       shutUp = FALSE;   /* true if we put a no echo command
				     * into the command file */
    Boolean	  errOff = FALSE;   /* true if we turned error checking
				     * off before printing the command
				     * and need to turn it back on */
    char       	  *cmdTemplate;	    /* Template to use when printing the
				     * command */
    char    	  *cmdStart;	    /* Start of expanded command */
    LstNode 	  cmdNode;  	    /* Node for replacing the command */

    noSpecials = (noExecute && ! (job->node->type & OP_MAKE));

    if (strcmp (cmd, "...") == 0) {
	if ((job->flags & JOB_IGNDOTS) == 0) {
	    job->tailCmds = Lst_Succ (Lst_Member (job->node->commands,
						  (ClientData)cmd));
	    return (1);
	}
	return (0);
    }

#define DBPRINTF(fmt, arg) if (DEBUG(JOB)) printf (fmt, arg); fprintf (job->cmdFILE, fmt, arg)

    numCommands += 1;

    /*
     * For debugging, we replace each command with the result of expanding
     * the variables in the command.
     */
    cmdNode = Lst_Member (job->node->commands, (ClientData)cmd);
    cmdStart = cmd = Var_Subst (cmd, job->node, FALSE);
    Lst_Replace (cmdNode, (ClientData)cmdStart);

    cmdTemplate = "%s\n";

    /*
     * Check for leading @' and -'s to control echoing and error checking.
     */
    while (*cmd == '@' || *cmd == '-') {
	if (*cmd == '@') {
	    shutUp = TRUE;
	} else {
	    errOff = TRUE;
	}
	cmd++;
    }

    if (shutUp) {
	if (! (job->flags & JOB_SILENT) && !noSpecials &&
	    commandShell->hasEchoCtl) {
		DBPRINTF ("%s\n", commandShell->echoOff);
	} else {
	    shutUp = FALSE;
	}
    }

    if (errOff) {
	if ( ! (job->flags & JOB_IGNERR) && !noSpecials) {
	    if (commandShell->hasErrCtl) {
		/*
		 * we don't want the error-control commands showing
		 * up either, so we turn off echoing while executing
		 * them. We could put another field in the shell
		 * structure to tell JobDoOutput to look for this
		 * string too, but why make it any more complex than
		 * it already is?
		 */
		if (! (job->flags & JOB_SILENT) && !shutUp &&
		    commandShell->hasEchoCtl) {
			DBPRINTF ("%s\n", commandShell->echoOff);
			DBPRINTF ("%s\n", commandShell->ignErr);
			DBPRINTF ("%s\n", commandShell->echoOn);
		} else {
		    DBPRINTF ("%s\n", commandShell->ignErr);
		}
	    } else if (commandShell->ignErr &&
		       (*commandShell->ignErr != '\0'))
	    {
		/*
		 * The shell has no error control, so we need to be
		 * weird to get it to ignore any errors from the command.
		 * If echoing is turned on, we turn it off and use the
		 * errCheck template to echo the command. Leave echoing
		 * off so the user doesn't see the weirdness we go through
		 * to ignore errors. Set cmdTemplate to use the weirdness
		 * instead of the simple "%s\n" template.
		 */
		if (! (job->flags & JOB_SILENT) && !shutUp &&
		    commandShell->hasEchoCtl) {
			DBPRINTF ("%s\n", commandShell->echoOff);
			DBPRINTF (commandShell->errCheck, cmd);
			shutUp = TRUE;
		}
		cmdTemplate = commandShell->ignErr;
		/*
		 * The error ignoration (hee hee) is already taken care
		 * of by the ignErr template, so pretend error checking
		 * is still on.
		 */
		errOff = FALSE;
	    } else {
		errOff = FALSE;
	    }
	} else {
	    errOff = FALSE;
	}
    }
    
    DBPRINTF (cmdTemplate, cmd);
    
    if (errOff) {
	/*
	 * If echoing is already off, there's no point in issuing the
	 * echoOff command. Otherwise we issue it and pretend it was on
	 * for the whole command...
	 */
	if (!shutUp && commandShell->hasEchoCtl) {
	    DBPRINTF ("%s\n", commandShell->echoOff);
	    shutUp = TRUE;
	}
	DBPRINTF ("%s\n", commandShell->errCheck);
    }
    if (shutUp) {
	DBPRINTF ("%s\n", commandShell->echoOn);
    }
    return (0);
}

/*-
 *-----------------------------------------------------------------------
 * JobSaveCommand --
 *	Save a command to be executed when everything else is done.
 *	Callback function for JobFinish...
 *
 * Results:
 *	Always returns 0
 *
 * Side Effects:
 *	The command is tacked onto the end of postCommands's commands list.
 *
 *-----------------------------------------------------------------------
 */
static int
JobSaveCommand (cmd, gn)
    char    *cmd;
    GNode   *gn;
{
    cmd = Var_Subst (cmd, gn, FALSE);
    (void)Lst_AtEnd (postCommands->commands, (ClientData)cmd);
    return (0);
}

/*-
 *-----------------------------------------------------------------------
 * JobFinish  --
 *	Do final processing for the given job including updating
 *	parents and starting new jobs as available/necessary. Note
 *	that we pay no attention to the JOB_IGNERR flag here.
 *	This is because when we're called because of a noexecute flag
 *	or something, jstat.w_status is 0 and when called from
 *	Job_CatchChildren, the status is zeroed if it s/b ignored.
 *
 * Results:
 *	None
 *
 * Side Effects:
 *	Some nodes may be put on the toBeMade queue.
 *	Final commands for the job are placed on postCommands.
 *-----------------------------------------------------------------------
 */
/*ARGSUSED*/
static void
JobFinish (job, status)
    Job           *job;	      	  /* job to finish */
    union wait	  status;     	  /* sub-why job went away */
{
    Boolean 	  done;

    if ((WIFEXITED(status) &&
	  (((status.w_retcode != 0) && !(job->flags & JOB_IGNERR)) ||
	   !backwards)) ||
	(WIFSIGNALED(status) && (status.w_termsig != SIGCONT)))
    {
	/*
	 * If it exited non-zero and either we're doing things our
	 * way or we're not ignoring errors, the job is finished.
	 * Similarly, if the shell died because of a signal (the
	 * conditional on SIGCONT is to handle the mapping of Sprite
	 * signal semantics whereby wait will return a signal
	 * termination with SIGCONT being the signal to indicate that the
	 * child has resumed), the job is also finished. In these
	 * cases, finish out the job's output before printing the exit
	 * status...
	 */
	if (usePipes) {
#ifdef CUSTOMS
	    Rpc_Ignore(job->inPipe);
#else
	    FD_CLR(job->inPipe, &outputs);
#endif /* CUSTOMS */
	    if (job->outPipe != job->inPipe) {
		(void)close (job->outPipe);
	    }
	    JobDoOutput (job, TRUE);
	    (void)close (job->inPipe);
	} else {
	    (void)close (job->outFd);
	    JobDoOutput (job, TRUE);
	}

	if (job->cmdFILE != NULL && job->cmdFILE != stdout) {
	    fclose(job->cmdFILE);
	}
	done = TRUE;
    } else {
	done = FALSE;
    }
    
    if (done ||
	WIFSTOPPED(status) ||
	(WIFSIGNALED(status) && (status.w_termsig == SIGCONT)) ||
	DEBUG(JOB))
    {
	FILE	  *out;
	
	if (backwards && !usePipes && (job->flags & JOB_IGNERR)) {
	    /*
	     * If output is going to a file and this job is ignoring
	     * errors, arrange to have the exit status sent to the
	     * output file as well.
	     */
	    out = fdopen (job->outFd, "w");
	} else {
	    out = stdout;
	}

	if (WIFEXITED(status)) {
	    if (status.w_retcode != 0) {
		if (usePipes && job->node != lastNode) {
		    fprintf (out, TARG_FMT, job->node->name);
		    lastNode = job->node;
		}
		fprintf (out, "*** Error code %d%s\n", status.w_retcode,
			 (job->flags & JOB_IGNERR) ? " (ignored)" : "");

		if (job->flags & JOB_IGNERR) {
		    status.w_status = 0;
		}
	    } else if (DEBUG(JOB)) {
		if (usePipes && job->node != lastNode) {
		    fprintf (out, TARG_FMT, job->node->name);
		    lastNode = job->node;
		}
		fprintf (out, "*** Completed successfully\n");
	    }
	} else if (WIFSTOPPED(status)) {
	    if (usePipes && job->node != lastNode) {
		fprintf (out, TARG_FMT, job->node->name);
		lastNode = job->node;
	    }
	    if (! (job->flags & JOB_REMIGRATE)) {
		fprintf (out, "*** Stopped -- signal %d\n", status.w_stopsig);
	    }
	    job->flags |= JOB_RESUME;
	    (void)Lst_AtEnd(stoppedJobs, (ClientData)job);
	    fflush(out);
	    return;
	} else if (status.w_termsig == SIGCONT) {
	    /*
	     * If the beastie has continued, shift the Job from the stopped
	     * list to the running one (or re-stop it if concurrency is
	     * exceeded) and go and get another child.
	     */
	    if (job->flags & (JOB_RESUME|JOB_REMIGRATE|JOB_RESTART)) {
		if (usePipes && job->node != lastNode) {
		    fprintf (out, TARG_FMT, job->node->name);
		    lastNode = job->node;
		}
		fprintf (out, "*** Continued\n");
	    }
	    if (! (job->flags & JOB_CONTINUING)) {
		JobRestart(job);
	    } else {
		Lst_AtEnd(jobs, (ClientData)job);
		nJobs += 1;
		if (! (job->flags & JOB_REMOTE)) {
		    nLocal += 1;
		}
		if (nJobs == maxJobs) {
		    jobFull = TRUE;
		    if (DEBUG(JOB)) {
			printf("Job queue is full.\n");
		    }
		}
	    }
	    fflush(out);
	    return;
	} else {
	    if (usePipes && job->node != lastNode) {
		fprintf (out, TARG_FMT, job->node->name);
		lastNode = job->node;
	    }
	    fprintf (out, "*** Signal %d\n", status.w_termsig);
	}

	fflush (out);
    }

    /*
     * Now handle the -B-mode stuff. If the beast still isn't finished,
     * try and restart the job on the next command. If JobStart says it's
     * ok, it's ok. If there's an error, this puppy is done.
     */
    if (backwards &&
	(status.w_status == 0) &&
	!Lst_IsAtEnd (job->node->commands))
    {
	switch (JobStart (job->node,
			  job->flags & JOB_IGNDOTS,
			  job))
	{
	    case JOB_RUNNING:
		done = FALSE;
		break;
	    case JOB_ERROR:
		done = TRUE;
		status.w_retcode = 1;
		break;
	    case JOB_FINISHED:
		/*
		 * If we got back a JOB_FINISHED code, JobStart has already
		 * called Make_Update and freed the job descriptor. We set
		 * done to false here to avoid fake cycles and double frees.
		 * JobStart needs to do the update so we can proceed up the
		 * graph when given the -n flag..
		 */
		done = FALSE;
		break;
	}
    } else {
	done = TRUE;
    }
		

    if (done &&
	(aborting != ABORT_ERROR) &&
	(aborting != ABORT_INTERRUPT) &&
	(status.w_status == 0))
    {
	/*
	 * As long as we aren't aborting and the job didn't return a non-zero
	 * status that we shouldn't ignore, we call Make_Update to update
	 * the parents. In addition, any saved commands for the node are placed
	 * on the .END target.
	 */
	if (job->tailCmds != NILLNODE) {
	    Lst_ForEachFrom (job->tailCmds, JobSaveCommand,
			     (ClientData)job->node);
	}
	Make_Update (job->node);
	free((Address)job);
    } else if (status.w_status) {
	errors += 1;
	free((Address)job);
    }

    while (!errors && !jobFull && !Lst_IsEmpty(stoppedJobs)) {
	JobRestart((Job *)Lst_DeQueue(stoppedJobs));
    }
}

/*-
 *-----------------------------------------------------------------------
 * Job_Touch --
 *	Touch the given target. Called by JobStart when the -t flag was
 *	given
 *
 * Results:
 *	None
 *
 * Side Effects:
 *	The data modification of the file is changed. In addition, if the
 *	file did not exist, it is created.
 *-----------------------------------------------------------------------
 */
void
Job_Touch (gn, silent)
    GNode         *gn;	      	/* the node of the file to touch */
    Boolean 	  silent;   	/* TRUE if should not print messages */
{
    int		  streamID;   	/* ID of stream opened to do the touch */
    struct timeval times[2];	/* Times for utimes() call */
    struct stat attr;        /* Attributes of the file */

    if (gn->type & (OP_JOIN|OP_USE|OP_EXEC|OP_DONTCARE)) {
	/*
	 * .JOIN, .USE, .ZEROTIME and .DONTCARE targets are "virtual" targets
	 * and, as such, shouldn't really be created.
	 */
	return;
    }
    
    if (!silent) {
	printf ("touch %s\n", gn->name);
    }

    if (noExecute) {
	return;
    }

    if (gn->type & OP_ARCHV) {
	Arch_Touch (gn);
    } else if (gn->type & OP_LIB) {
	Arch_TouchLib (gn);
    } else {
	char	*file = gn->path ? gn->path : gn->name;

	times[0].tv_sec = times[1].tv_sec = now;
	times[0].tv_usec = times[1].tv_usec = 0;
	if (utimes(file, times) < 0){
	    streamID = open (file, O_RDWR | O_CREAT, 0666);

	    if (streamID >= 0) {
		char	c;

		/*
		 * Read and write a byte to the file to change the
		 * modification time, then close the file.
		 */
		if (read(streamID, &c, 1) == 1) {
		    lseek(streamID, 0L, L_SET);
		    write(streamID, &c, 1);
		}
		
		(void)close (streamID);
	    } else {
		extern char *sys_errlist[];

		printf("*** couldn't touch %s: %s", file, sys_errlist[errno]);
	    }
	}
    }
}

/*-
 *-----------------------------------------------------------------------
 * Job_CheckCommands --
 *	Make sure the given node has all the commands it needs. 
 *
 * Results:
 *	TRUE if the commands list is/was ok.
 *
 * Side Effects:
 *	The node will have commands from the .DEFAULT rule added to it
 *	if it needs them.
 *-----------------------------------------------------------------------
 */
Boolean
Job_CheckCommands (gn, abortProc)
    GNode          *gn;	    	    /* The target whose commands need
				     * verifying */
    void    	  (*abortProc)();   /* Function to abort with message */
{
    if (OP_NOP(gn->type) && Lst_IsEmpty (gn->commands) &&
	(gn->type & OP_LIB) == 0) {
	/*
	 * No commands. Look for .DEFAULT rule from which we might infer
	 * commands 
	 */
	if ((DEFAULT != NILGNODE) && !Lst_IsEmpty(DEFAULT->commands)) {
	    /*
	     * Make only looks for a .DEFAULT if the node was never the
	     * target of an operator, so that's what we do too If
	     * a .DEFAULT was given, we substitute its commands for gn's
	     * commands and set the IMPSRC variable to be the target's name 
	     */
	    Lst_Destroy (gn->commands, NOFREE);
	    gn->commands = DEFAULT->commands;
	    Var_Set (IMPSRC, Var_Value (TARGET, gn), gn);
	} else if (Dir_MTime (gn) == 0) {
	    /*
	     * The node wasn't the target of an operator we have no .DEFAULT
	     * rule to go on and the target doesn't already exist. There's
	     * nothing more we can do for this branch. If the -k flag wasn't
	     * given, we stop in our tracks, otherwise we just don't update
	     * this node's parents so they never get examined. 
	     */
	    if (gn->type & OP_DONTCARE) {
		printf ("Can't figure out how to make %s (ignored)\n",
			gn->name);
	    } else if (keepgoing) {
		printf ("Can't figure out how to make %s (continuing)\n",
			gn->name);
		return (FALSE);
	    } else {
		(*abortProc) ("Can't figure out how to make %s. Stop",
			     gn->name);
		/*NOTREACHED*/
	    }
	}
    }
    return (TRUE);
}

/*-
 *-----------------------------------------------------------------------
 * JobExec --
 *	Execute the shell for the given job. Called from JobStart and
 *	JobRestart.
 *
 * Results:
 *	None.
 *
 * Side Effects:
 *	A shell is executed, outputs is altered and the Job structure added
 *	to the job table.
 *
 *-----------------------------------------------------------------------
 */
static void
JobExec(job, argv)
    Job	    	  *job; 	/* Job to execute */
    char    	  **argv;
{
    int	    	  cpid;	    	/* ID of new child */
    
    if (DEBUG(JOB)) {
	int 	  i;
	
	printf("Running %s %sly\n", job->node->name,
	       job->flags&JOB_REMOTE?"remote":"local");
	printf("\tCommand: ");
	for (i = 0; argv[i] != (char *)NULL; i++) {
	    printf("%s ", argv[i]);
	}
	printf("\n");
    }
    
    /*
     * Some jobs produce no output and it's disconcerting to have
     * no feedback of their running (since they produce no output, the
     * banner with their name in it never appears). This is an attempt to
     * provide that feedback, even if nothing follows it.
     */
    if ((lastNode != job->node) && (job->flags & JOB_FIRST)) {
	printf(TARG_FMT, job->node->name);
	lastNode = job->node;
    }
    
#ifdef CUSTOMS
    if (job->flags & JOB_REMOTE) {
	job->cmdFd = fileno(job->cmdFILE);
	lseek(job->cmdFd, 0, L_SET);
	job->curPos = 0;
	Rpc_Watch(job->rmtFd, RPC_READABLE|RPC_WRITABLE, JobRmtIO,
		  (Rpc_Opaque)job);
	/*
	 * Now the job is actually running, add it to the table.
	 */
	nJobs += 1;
	(void)Lst_AtEnd (jobs, (ClientData)job);
	if (nJobs == maxJobs) {
	    jobFull = TRUE;
	    if (DEBUG(JOB)) {
		printf("Job queue is full.\n");
	    }
	}
    } else 
#endif CUSTOMS	
    if ((cpid =  vfork()) == -1) {
	Punt ("Cannot fork");
    } else if (cpid == 0) {

	/*
	 * Must duplicate the input stream down to the child's input and
	 * reset it to the beginning (again). Since the stream was marked
	 * close-on-exec, we must clear that bit in the new input.
	 */
	(void) dup2(fileno(job->cmdFILE), 0);
	fcntl(0, F_SETFD, 0);
	lseek(0, 0, L_SET);
	
	if (usePipes) {
	    /*
	     * Set up the child's output to be routed through the pipe
	     * we've created for it.
	     */
	    (void) dup2 (job->outPipe, 1);
	} else {
	    /*
	     * We're capturing output in a file, so we duplicate the
	     * descriptor to the temporary file into the standard
	     * output.
	     */
	    (void) dup2 (job->outFd, 1);
	}
	/*
	 * The output channels are marked close on exec. This bit was
	 * duplicated by the dup2 (on some systems), so we have to clear
	 * it before routing the shell's error output to the same place as
	 * its standard output.
	 */
	fcntl(1, F_SETFD, 0);
	(void) dup2 (1, 2);

#ifdef USE_PGRP
	/*
	 * We want to switch the child into a different process family so
	 * we can kill it and all its descendants in one fell swoop,
	 * by killing its process family, but not commit suicide.
	 */
	
#if defined(SYSV)
	(void) setpgrp();
#else
	(void) setpgrp(0, getpid());
#endif
#endif USE_PGRP

	if (job->flags & JOB_REMOTE) {
	    Rmt_Exec (shellPath, argv, FALSE);
	} else {
	    (void) execv (shellPath, argv);
	}

	(void) write (2, "Could not execute shell\n",
		 sizeof ("Could not execute shell"));
	_exit (1);
    } else {
	job->pid = cpid;

	/*
	 * Now the job is actually running, add it to the table.
	 */
	nJobs += 1;
	(void)Lst_AtEnd (jobs, (ClientData)job);
	if (nJobs == maxJobs) {
	    jobFull = TRUE;
	}
	
	if (usePipes && (job->flags & JOB_FIRST) ) {
	    /*
	     * The first time a job is run for a node, we set the current
	     * position in the buffer to the beginning and mark another
	     * stream to watch in the outputs mask
	     */
	    job->curPos = 0;
	    
#ifdef CUSTOMS
	    Rpc_Watch(job->inPipe, RPC_READABLE, JobLocalInput,
		      (Rpc_Opaque)job);
#else
	    FD_SET(job->inPipe, &outputs);
#endif /* CUSTOMS */
	}

	if (job->flags & JOB_REMOTE) {
	    job->rmtID = Rmt_LastID(job->pid);
	} else {
	    nLocal += 1;
#ifndef CUSTOMS
	    if (job->cmdFILE != stdout) {
		fclose(job->cmdFILE);
		job->cmdFILE = NULL;
	    }
#endif /* CUSTOMS */
	}
    }
}

/*-
 *-----------------------------------------------------------------------
 * JobMakeArgv --
 *	Create the argv needed to execute the shell for a given job.
 *	
 *
 * Results:
 *
 * Side Effects:
 *
 *-----------------------------------------------------------------------
 */
static void
JobMakeArgv(job, argv)
    Job	    	  *job;
    char	  **argv;
{
    int	    	  argc;
    static char	  args[10]; 	/* For merged arguments */
    
    argv[0] = shellName;
    argc = 1;

    if ((commandShell->exit && (*commandShell->exit != '-')) ||
	(commandShell->echo && (*commandShell->echo != '-')))
    {
	/*
	 * At least one of the flags doesn't have a minus before it, so
	 * merge them together. Have to do this because the *(&(@*#*&#$#
	 * Bourne shell thinks its second argument is a file to source.
	 * Grrrr. Note the ten-character limitation on the combined arguments.
	 */
	(void)sprintf(args, "-%s%s",
		      ((job->flags & JOB_IGNERR) ? "" :
		       (commandShell->exit ? commandShell->exit : "")),
		      ((job->flags & JOB_SILENT) ? "" :
		       (commandShell->echo ? commandShell->echo : "")));

	if (args[1]) {
	    argv[argc] = args;
	    argc++;
	}
    } else {
	if (!(job->flags & JOB_IGNERR) && commandShell->exit) {
	    argv[argc] = commandShell->exit;
	    argc++;
	}
	if (!(job->flags & JOB_SILENT) && commandShell->echo) {
	    argv[argc] = commandShell->echo;
	    argc++;
	}
    }
    argv[argc] = (char *)NULL;
}

/*-
 *-----------------------------------------------------------------------
 * JobRestart --
 *	Restart a job that stopped for some reason. If the job stopped
 *	because it migrated home again, we tell the Rmt module to
 *	find a new home for it and make it runnable if Rmt_ReExport
 *	succeeded (if it didn't and the job may be run locally, we
 *	simply resume it). If the job didn't run and can now, we run it.
 *
 * Results:
 *	None.
 *
 * Side Effects:
 *	jobFull will be set if the job couldn't be run.
 *
 *-----------------------------------------------------------------------
 */
static void
JobRestart(job)
    Job 	  *job;    	/* Job to restart */
{
    if (job->flags & JOB_REMIGRATE) {
	if (DEBUG(JOB)) {
	    printf("Remigrating %x\n", job->pid);
	}
	if (!Rmt_ReExport(job->pid)) {
	    if (DEBUG(JOB)) {
		printf("Couldn't migrate...");
	    }
	    if (nLocal != maxLocal) {
		/*
		 * Job cannot be remigrated, but there's room on the local
		 * machine, so resume the job and note that another
		 * local job has started.
		 */
		if (DEBUG(JOB)) {
		    printf("resuming on local machine\n");
	        }
#ifdef USE_PGRP
		killpg(job->pid, SIGCONT);
#else
		kill(job->pid, SIGCONT);
#endif /* USE_PGRP */
		nLocal +=1;
		job->flags &= ~(JOB_REMIGRATE|JOB_RESUME);
	    } else {
		/*
		 * Job cannot be restarted. Mark the table as full and
		 * place the job back on the list of stopped jobs.
		 */
		if (DEBUG(JOB)) {
		    printf("holding\n");
		}
		(void)Lst_AtFront(stoppedJobs, (ClientData)job);
		jobFull = TRUE;
		if (DEBUG(JOB)) {
		    printf("Job queue is full.\n");
		}
		return;
	    }
	} else {
	    /*
	     * Clear out the remigrate and resume flags. If MIGRATE was set,
	     * leave that around for JobFinish to see so it doesn't print out
	     * that the job was continued.
	     */
	    job->flags &= ~(JOB_REMIGRATE|JOB_RESUME);
	}
	
	(void)Lst_AtEnd(jobs, (ClientData)job);
	nJobs += 1;
	if (nJobs == maxJobs) {
	    jobFull = TRUE;
	    if (DEBUG(JOB)) {
		printf("Job queue is full.\n");
	    }
	}
    } else if (job->flags & JOB_RESTART) {
	/*
	 * Set up the control arguments to the shell. This is based on the
	 * flags set earlier for this job. If the JOB_IGNERR flag is clear,
	 * the 'exit' flag of the commandShell is used to cause it to exit
	 * upon receiving an error. If the JOB_SILENT flag is clear, the
	 * 'echo' flag of the commandShell is used to get it to start echoing
	 * as soon as it starts processing commands. 
	 */
	char	  *argv[4];
	
	JobMakeArgv(job, argv);
	
	if (DEBUG(JOB)) {
	    printf("Restarting %s...", job->node->name);
	}
#ifdef CUSTOMS
	if ((job->node->type&OP_NOEXPORT) || !JobExport(shellPath, argv, job))
#else
	if ((job->node->type&OP_NOEXPORT) || !Rmt_Begin(shellPath,
							argv,
							job->node))
#endif 
	{
	    if ((nLocal >= maxLocal) && ! (job->flags & JOB_SPECIAL)) {
		/*
		 * Can't be exported and not allowed to run locally -- put it
		 * back on the hold queue and mark the table full
		 */
		if (DEBUG(JOB)) {
		    printf("holding\n");
		}
		(void)Lst_AtFront(stoppedJobs, (ClientData)job);
		jobFull = TRUE;
		if (DEBUG(JOB)) {
		    printf("Job queue is full.\n");
		}
		return;
	    } else {
		/*
		 * Job may be run locally.
		 */
		if (DEBUG(JOB)) {
		    printf("running locally\n");
		}
		job->flags &= ~JOB_REMOTE;
	    }
	} else {
	    /*
	     * Can be exported. Hooray!
	     */
	    if (DEBUG(JOB)) {
		printf("exporting\n");
	    }
	    job->flags |= JOB_REMOTE;
	}
	JobExec(job, argv);
    } else {
	/*
	 * The job has stopped and needs to be restarted. Why it stopped,
	 * we don't know...
	 */
	if (DEBUG(JOB)) {
	    printf("Resuming %s...", job->node->name);
	}
	if (((job->flags & JOB_REMOTE) ||
	     (nLocal < maxLocal) ||
	     (((job->flags & JOB_SPECIAL) ||
	       (job->node->type & OP_NOEXPORT)) &&
	      (maxLocal == 0))) &&
	    (nJobs != maxJobs))
	{
	    /*
	     * If the job is remote, it's ok to resume it as long as the
	     * maximum concurrency won't be exceeded. If it's local and
	     * we haven't reached the local concurrency limit already (or the
	     * job must be run locally and maxLocal is 0), it's also ok to
	     * resume it.
	     */
	    Boolean error;
	    extern int errno;
	    extern char *sys_errlist[];
	    union wait status;
	    
#ifdef CUSTOMS
	    if (job->flags & JOB_REMOTE) {
		(void)PassSigIfRemote(job, SIGCONT);
		error = 0;
	    }
#else
# ifdef USE_PGRP
	    error = (killpg(job->pid, SIGCONT) != 0);
# else
	    error = (kill(job->pid, SIGCONT) != 0);
# endif /* USE_PGRP */
#endif /* CUSTOMS */

	    if (!error) {
		/*
		 * Make sure the user knows we've continued the beast and
		 * actually put the thing in the job table.
		 */
		job->flags |= JOB_CONTINUING;
		status.w_termsig = SIGCONT;
		JobFinish(job, status);
		
		job->flags &= ~(JOB_RESUME|JOB_CONTINUING);
		if (DEBUG(JOB)) {
		    printf("done\n");
		}
	    } else {
		Error("couldn't resume %s: %s", job->node->name,
		      sys_errlist[errno]);
		status.w_status = 0;
		status.w_retcode = 1;
		JobFinish(job, status);
	    }
	} else {
	    /*
	     * Job cannot be restarted. Mark the table as full and
	     * place the job back on the list of stopped jobs. We're called
	     * if a job is already running of its own accord, so send the
	     * puppy a stop signal it can't ignore just to make sure the
	     * concurrency limit isn't exceeded.
	     */
	    if (DEBUG(JOB)) {
		printf("table full\n");
	    }
#ifndef CUSTOMS
# ifdef USE_PGRP
	    killpg(job->pid, SIGSTOP);
# else
	    kill(job->pid, SIGSTOP);
# endif /* USE_PGRP */
#endif /* CUSTOMS */
	    (void)Lst_AtFront(stoppedJobs, (ClientData)job);
	    jobFull = TRUE;
	    if (DEBUG(JOB)) {
		printf("Job queue is full.\n");
	    }
	}
    }
}

/*-
 *-----------------------------------------------------------------------
 * JobStart  --
 *	Start a target-creation process going for the target described
 *	by the graph node gn. 
 *
 * Results:
 *	JOB_ERROR if there was an error in the commands, JOB_FINISHED
 *	if there isn't actually anything left to do for the job and
 *	JOB_RUNNING if the job has been started.
 *
 * Side Effects:
 *	A new Job node is created and added to the list of running
 *	jobs. PMake is forked and a child shell created.
 *-----------------------------------------------------------------------
 */
static int
JobStart (gn, flags, previous)
    GNode         *gn;	      /* target to create */
    short	  flags;      /* flags for the job to override normal ones.
			       * e.g. JOB_SPECIAL or JOB_IGNDOTS */
    Job 	  *previous;  /* The previous Job structure for this node,
			       * if any. */
{
    register Job  *job;       /* new job descriptor */
    char	  *argv[4];   /* Argument vector to shell */
    char          args[5];    /* arguments to shell */
    static int    jobno = 0;  /* job number of catching output in a file */
    Boolean	  cmdsOK;     /* true if the nodes commands were all right */
    Boolean 	  local;      /* Set true if the job was run locally */
    Boolean 	  noExec;     /* Set true if we decide not to run the job */

    if (previous != (Job *)NULL) {
	previous->flags &= ~ (JOB_FIRST|JOB_IGNERR|JOB_SILENT|JOB_REMOTE);
	job = previous;
    } else {
	job = (Job *) malloc (sizeof (Job));
	if (job == (Job *)NULL) {
	    Punt("JobStart out of memory");
	}
	flags |= JOB_FIRST;
    }

    job->node = gn;
    job->tailCmds = NILLNODE;

    /*
     * Set the initial value of the flags for this job based on the global
     * ones and the node's attributes... Any flags supplied by the caller
     * are also added to the field.
     */
    job->flags = 0;
    if (Targ_Ignore (gn)) {
	job->flags |= JOB_IGNERR;
    }
    if (Targ_Silent (gn)) {
	job->flags |= JOB_SILENT;
    }
    job->flags |= flags;

    /*
     * If the -n flag wasn't given, we open up OUR (not the child's)
     * temporary file to stuff commands in it. The thing is rd/wr so we don't
     * need to reopen it to feed it to the shell. If the -n flag *was* given,
     * we just set the file to be stdout. Cute, huh?
     */
    if ((gn->type & OP_MAKE) || (!noExecute && !touchFlag)) {
	job->cmdFILE = fopen (tfile, "w+");
	if (job->cmdFILE == (FILE *) NULL) {
	    Punt ("Could not open %s", tfile);
	}
	fcntl(fileno(job->cmdFILE), F_SETFD, 1);
	/*
	 * Send the commands to the command file, flush all its buffers then
	 * rewind and remove the thing.
	 */
	noExec = FALSE;

	if (backwards) {
	    /*
	     * Be compatible: If this is the first time for this node,
	     * verify its commands are ok and open the commands list for
	     * sequential access by later invocations of JobStart.
	     * Once that is done, we take the next command off the list
	     * and print it to the command file. If the command was an
	     * ellipsis, note that there's nothing more to execute.
	     */
	    if (job->flags & JOB_FIRST) {
		cmdsOK = Job_CheckCommands (gn, Punt);
		if (Lst_Open (gn->commands) != SUCCESS) {
		    cmdsOK = FALSE;
		}
	    } else {
		cmdsOK = TRUE;
	    }
	    
	    if (cmdsOK) {
		LstNode	ln = Lst_Next (gn->commands);
		    
		if ((ln == NILLNODE) ||
		    JobPrintCommand ((char *)Lst_Datum (ln), job))
		{
		    noExec = TRUE;
		    Lst_Close (gn->commands);
		}
		if (noExec && !(job->flags & JOB_FIRST)) {
		    /*
		     * If we're not going to execute anything, the job
		     * is done and we need to close down the various
		     * file descriptors we've opened for output, then
		     * call JobDoOutput to catch the final characters or
		     * send the file to the screen... Note that the i/o streams
		     * are only open if this isn't the first job.
		     * Note also that this could not be done in
		     * Job_CatchChildren b/c it wasn't clear if there were
		     * more commands to execute or not...
		     */
		    if (usePipes) {
#ifdef CUSTOMS
			Rpc_Ignore(job->inPipe);
#else
			FD_CLR(job->inPipe, &outputs);
#endif				/* CUSTOMS */
			if (job->outPipe != job->inPipe) {
			    (void)close (job->outPipe);
			}
			JobDoOutput (job, TRUE);
			(void)close (job->inPipe);
		    } else {
			(void)close (job->outFd);
			JobDoOutput (job, TRUE);
		    }
		}
	    }
	} else {
	    /*
	     * We can do all the commands at once. hooray for sanity
	     */
	    numCommands = 0;
	    cmdsOK = Job_CheckCommands (gn, Punt);
	    Lst_ForEach (gn->commands, JobPrintCommand, (ClientData)job);
	    
	    /*
	     * If we didn't print out any commands to the shell script,
	     * there's not much point in executing the shell, is there?
	     */
	    if (numCommands == 0) {
		noExec = TRUE;
	    }
	}
    } else if (noExecute) {
	/*
	 * Not executing anything -- just print all the commands to stdout
	 * in one fell swoop. This will still set up job->tailCmds correctly.
	 */
	if (lastNode != gn) {
	    printf (TARG_FMT, gn->name);
	    lastNode = gn;
	}
	job->cmdFILE = stdout;
	/*
	 * Only print the commands if they're ok, but don't die if they're
	 * not -- just let the user know they're bad and keep going. It
	 * doesn't do any harm in this case and may do some good.
	 */
	if (Job_CheckCommands(gn, Error)) {
	    Lst_ForEach(gn->commands, JobPrintCommand, (ClientData)job);
	}
	/*
	 * Don't execute the shell, thank you.
	 */
	noExec = TRUE;
    } else {
	/*
	 * Just touch the target and note that no shell should be executed.
	 * Set cmdFILE to stdout to make life easier. Check the commands, too,
	 * but don't die if they're no good -- it does no harm to keep working
	 * up the graph.
	 */
	cmdsOK = Job_CheckCommands(gn, Error);
	job->cmdFILE = stdout;
    	Job_Touch (gn, job->flags&JOB_SILENT);
	noExec = TRUE;
    }

    /*
     * If we're not supposed to execute a shell, don't. 
     */
    if (noExec) {
	/*
	 * Unlink and close the command file if we opened one
	 */
	if (job->cmdFILE != stdout) {
	    (void) unlink (tfile);
	    fclose(job->cmdFILE);
	} else {
	    fflush (stdout);
	}

	/*
	 * We only want to work our way up the graph if we aren't here because
	 * the commands for the job were no good.
	 */
	if (cmdsOK) {
	    if (aborting == 0) {
		if (job->tailCmds != NILLNODE) {
		    Lst_ForEachFrom(job->tailCmds, JobSaveCommand,
				    (ClientData)job->node);
		}
		Make_Update(job->node);
	    }
	    free((Address)job);
	    return(JOB_FINISHED);
	} else {
	    free((Address)job);
	    return(JOB_ERROR);
	}
    } else {
	fflush (job->cmdFILE);
	(void) unlink (tfile);
    }

    /*
     * Set up the control arguments to the shell. This is based on the flags
     * set earlier for this job.
     */
    JobMakeArgv(job, argv);

    /*
     * If we're using pipes to catch output, create the pipe by which we'll
     * get the shell's output. If we're using files, print out that we're
     * starting a job and then set up its temporary-file name. This is just
     * tfile with two extra digits tacked on -- jobno.
     */
    if (!backwards || (job->flags & JOB_FIRST)) {
	if (usePipes) {
	    int fd[2];
	    (void) pipe(fd); 
	    job->inPipe = fd[0];
	    job->outPipe = fd[1];
	    (void)fcntl (job->inPipe, F_SETFD, 1);
	    (void)fcntl (job->outPipe, F_SETFD, 1);
	} else {
	    printf ("Remaking `%s'\n", gn->name);
	    fflush (stdout);
	    sprintf (job->outFile, "%s%02d", tfile, jobno);
	    jobno = (jobno + 1) % 100;
	    job->outFd = open(job->outFile,O_WRONLY|O_CREAT|O_APPEND,0600);
	    (void)fcntl (job->outFd, F_SETFD, 1);
	}
    }

    if (!(gn->type & OP_NOEXPORT)) {
#ifdef CUSTOMS
	local = !JobExport(shellPath, argv, job);
#else
	local = !Rmt_Begin (shellPath, argv, gn);
#endif /* CUSTOMS */
	if (!local) {
	    job->flags |= JOB_REMOTE;
	}
    } else {
	local = TRUE;
    }

    if (local && (nLocal >= maxLocal) &&
	! ((job->flags & JOB_SPECIAL) ||
	    ((gn->type & OP_NOEXPORT) && (maxLocal == 0))))
    {
	/*
	 * The job can only be run locally, but we've hit the limit of
	 * local concurrency, so put the job on hold until some other job
	 * finishes. Note that the special jobs (.BEGIN, .INTERRUPT and .END)
	 * may be run locally even when the local limit has been reached
	 * (e.g. when maxLocal == 0), though they will be exported if at
	 * all possible. In addition, any target marked with .NOEXPORT will
	 * be run locally if maxLocal is 0.
	 */
	jobFull = TRUE;
	
	if (DEBUG(JOB)) {
	    printf("Can only run job locally.\n");
	}
	job->flags |= JOB_RESTART;
	(void)Lst_AtEnd(stoppedJobs, (ClientData)job);
    } else {
	if ((nLocal >= maxLocal) && local) {
	    /*
	     * If we're running this job locally as a special case (see above),
	     * at least say the table is full.
	     */
	    jobFull = TRUE;
	    if (DEBUG(JOB)) {
		printf("Local job queue is full.\n");
	    }
	}
	JobExec(job, argv);
    }
    return(JOB_RUNNING);
}

/*-
 *-----------------------------------------------------------------------
 * JobDoOutput  --
 *	This function is called at different times depending on
 *	whether the user has specified that output is to be collected
 *	via pipes or temporary files. In the former case, we are called
 *	whenever there is something to read on the pipe. We collect more
 *	output from the given job and store it in the job's outBuf. If
 *	this makes up a line, we print it tagged by the job's identifier,
 *	as necessary.
 *	If output has been collected in a temporary file, we open the
 *	file and read it line by line, transfering it to our own
 *	output channel until the file is empty. At which point we
 *	remove the temporary file.
 *	In both cases, however, we keep our figurative eye out for the
 *	'noPrint' line for the shell from which the output came. If
 *	we recognize a line, we don't print it. If the command is not
 *	alone on the line (the character after it is not \0 or \n), we
 *	do print whatever follows it.
 *
 * Results:
 *	None
 *
 * Side Effects:
 *	curPos may be shifted as may the contents of outBuf.
 *-----------------------------------------------------------------------
 */
static void
JobDoOutput (job, finish)
    register Job   *job;	  /* the job whose output needs printing */
    Boolean	   finish;	  /* TRUE if this is the last time we'll be
				   * called for this job */
{
    Boolean       gotNL = FALSE;  /* true if got a newline */
    register int  nr;	      	  /* number of bytes read */
    register int  i;	      	  /* auxiliary index into outBuf */
    register int  max;	      	  /* limit for i (end of current data) */
    int		  nRead;      	  /* (Temporary) number of bytes read */

    char          c;		  /* character after noPrint string */
    FILE      	  *oFILE;	  /* Stream pointer to shell's output file */
    char          inLine[132];

    
    if (usePipes) {
	/*
	 * Read as many bytes as will fit in the buffer.
	 */
end_loop:
	
	nRead = read (job->inPipe, &job->outBuf[job->curPos],
			 JOB_BUFSIZE - job->curPos);
	if (nRead < 0) {
	    if (DEBUG(JOB)) {
		perror("JobDoOutput(piperead)");
	    }
	    nr = 0;
	} else {
	    nr = nRead;
	}

	/*
	 * If we hit the end-of-file (the job is dead), we must flush its
	 * remaining output, so pretend we read a newline if there's any
	 * output remaining in the buffer.
	 * Also clear the 'finish' flag so we stop looping.
	 */
	if ((nr == 0) && (job->curPos != 0)) {
	    job->outBuf[job->curPos] = '\n';
	    nr = 1;
	    finish = FALSE;
	} else if (nr == 0) {
	    finish = FALSE;
	}
	
	/*
	 * Look for the last newline in the bytes we just got. If there is
	 * one, break out of the loop with 'i' as its index and gotNL set
	 * TRUE. 
	 */
	max = job->curPos + nr;
	for (i = job->curPos + nr - 1; i >= job->curPos; i--) {
	    if (job->outBuf[i] == '\n') {
		gotNL = TRUE;
		break;
	    } else if (job->outBuf[i] == '\0') {
		/*
		 * Why?
		 */
		job->outBuf[i] = ' ';
	    }
	}
	
	if (!gotNL) {
	    job->curPos += nr;
	    if (job->curPos == JOB_BUFSIZE) {
		/*
		 * If we've run out of buffer space, we have no choice
		 * but to print the stuff. sigh. 
		 */
		gotNL = TRUE;
		i = job->curPos;
	    }
	}
	if (gotNL) {
	    /*
	     * Need to send the output to the screen. Null terminate it
	     * first, overwriting the newline character if there was one.
	     * So long as the line isn't one we should filter (according
	     * to the shell description), we print the line, preceeded
	     * by a target banner if this target isn't the same as the
	     * one for which we last printed something.
	     * The rest of the data in the buffer are then shifted down
	     * to the start of the buffer and curPos is set accordingly. 
	     */
	    job->outBuf[i] = '\0';
	    if (i >= job->curPos) {
		register char	*cp, *ecp;

		cp = job->outBuf;
		if (commandShell->noPrint) {
		    ecp = Str_FindSubstring(job->outBuf,
					    commandShell->noPrint);
		    while (ecp != (char *)NULL) {
			if (cp != ecp) {
			    *ecp = '\0';
			    if (job->node != lastNode) {
				printf (TARG_FMT, job->node->name);
				lastNode = job->node;
			    }
			    /*
			     * The only way there wouldn't be a newline after
			     * this line is if it were the last in the buffer.
			     * however, since the non-printable comes after it,
			     * there must be a newline, so we don't print one.
			     */
			    printf ("%s", cp);
			}
			cp = ecp + commandShell->noPLen;
			if (cp != &job->outBuf[i]) {
			    /*
			     * Still more to print, look again after skipping
			     * the whitespace following the non-printable
			     * command....
			     */
			    cp++;
			    while (*cp == ' ' || *cp == '\t' || *cp == '\n') {
				cp++;
			    }
			    ecp = Str_FindSubstring (cp,
						     commandShell->noPrint);
			} else {
			    break;
			}
		    }
		}

		/*
		 * There's still more in that thar buffer. This time, though,
		 * we know there's no newline at the end, so we add one of
		 * our own free will.
		 */
		if (*cp != '\0') {
		    if (job->node != lastNode) {
			printf (TARG_FMT, job->node->name);
			lastNode = job->node;
		    }
		    printf ("%s\n", cp);
		}

		fflush (stdout);
	    }
	    if (i < max - 1) {
		bcopy (&job->outBuf[i + 1], /* shift the remaining */
		       job->outBuf,        /* characters down */
		       max - (i + 1));
		job->curPos = max - (i + 1);
		
	    } else {
		/*
		 * We have written everything out, so we just start over
		 * from the start of the buffer. No copying. No nothing.
		 */
		job->curPos = 0;
	    }
	}
	if (finish) {
	    /*
	     * If the finish flag is true, we must loop until we hit
	     * end-of-file on the pipe. This is guaranteed to happen eventually
	     * since the other end of the pipe is now closed (we closed it
	     * explicitly and the child has exited). When we do get an EOF,
	     * finish will be set FALSE and we'll fall through and out.
	     */
	    goto end_loop;
	}
    } else {
	/*
	 * We've been called to retrieve the output of the job from the
	 * temporary file where it's been squirreled away. This consists of
	 * opening the file, reading the output line by line, being sure not
	 * to print the noPrint line for the shell we used, then close and
	 * remove the temporary file. Very simple.
	 *
	 * Change to read in blocks and do FindSubString type things as for
	 * pipes? That would allow for "@echo -n..."
	 */
	oFILE = fopen (job->outFile, "r");
	if (oFILE != (FILE *) NULL) {
	    printf ("Results of making %s:\n", job->node->name);
	    while (fgets (inLine, sizeof(inLine), oFILE) != NULL) {
		register char	*cp, *ecp, *endp;

		cp = inLine;
		endp = inLine + strlen(inLine);
		if (endp[-1] == '\n') {
		    *--endp = '\0';
		}
		if (commandShell->noPrint) {
		    ecp = Str_FindSubstring(cp, commandShell->noPrint);
		    while (ecp != (char *)NULL) {
			if (cp != ecp) {
			    *ecp = '\0';
			    /*
			     * The only way there wouldn't be a newline after
			     * this line is if it were the last in the buffer.
			     * however, since the non-printable comes after it,
			     * there must be a newline, so we don't print one.
			     */
			    printf ("%s", cp);
			}
			cp = ecp + commandShell->noPLen;
			if (cp != endp) {
			    /*
			     * Still more to print, look again after skipping
			     * the whitespace following the non-printable
			     * command....
			     */
			    cp++;
			    while (*cp == ' ' || *cp == '\t' || *cp == '\n') {
				cp++;
			    }
			    ecp = Str_FindSubstring(cp, commandShell->noPrint);
			} else {
			    break;
			}
		    }
		}

		/*
		 * There's still more in that thar buffer. This time, though,
		 * we know there's no newline at the end, so we add one of
		 * our own free will.
		 */
		if (*cp != '\0') {
		    printf ("%s\n", cp);
		}
	    }
	    fclose (oFILE);
	    (void) unlink (job->outFile);
	}
    }
    fflush(stdout);
}

/*-
 *-----------------------------------------------------------------------
 * Job_CatchChildren --
 *	Handle the exit of a child. Called from Make_Make.
 *
 * Results:
 *	none.
 *
 * Side Effects:
 *	The job descriptor is removed from the list of children.
 *
 * Notes:
 *	We do waits, blocking or not, according to the wisdom of our
 *	caller, until there are no more children to report. For each
 *	job, call JobFinish to finish things off. This will take care of
 *	putting jobs on the stoppedJobs queue.
 *
 *	If we got an error and are aborting (aborting == ABORT_ERROR) and
 *	the job list is now empty, we are done for the day.
 *	If we recognized an error (errors !=0), we set the aborting flag
 *	to ABORT_ERROR so no more jobs will be started.
 *-----------------------------------------------------------------------
 */
void
Job_CatchChildren (block)
    Boolean	  block;    	/* TRUE if should block on the wait. */
{
    int    	  pid;	    	/* pid of dead child */
    register Job  *job;	    	/* job descriptor for dead child */
    LstNode       jnode;    	/* list element for finding job */
    union wait	  status;   	/* Exit/termination status */

    while ((pid = wait3(&status, (block?0:WNOHANG)|WUNTRACED,
			(struct rusage *)0)) > 0)
    {
	if (DEBUG(JOB)) {
#ifdef Sprite
	    printf("Process %x exited or stopped.\n", pid);
#else
	    printf("Process %d exited or stopped.\n", pid);
#endif /* Sprite */
	}
	    

	jnode = Lst_Find (jobs, (ClientData)pid, JobCmpPid);

	if (jnode == NILLNODE) {
	    if (WIFSIGNALED(status) && (status.w_termsig == SIGCONT)) {
		jnode = Lst_Find(stoppedJobs, (ClientData)pid, JobCmpPid);
		if (jnode == NILLNODE) {
#ifdef Sprite
		    Error("Resumed child (%x) not in table", pid);
#else
		    Error("Resumed child (%d) not in table", pid);
#endif /* Sprite */
		    continue;
		}
		job = (Job *)Lst_Datum(jnode);
		(void)Lst_Remove(stoppedJobs, jnode);
	    } else {
#ifdef Sprite
		Error ("Child (%x) not in table?", pid);
#else
		Error ("Child (%d) not in table?", pid);
#endif /* Sprite */
		continue;
	    }
	} else {
	    job = (Job *) Lst_Datum (jnode);
	    (void)Lst_Remove (jobs, jnode);
	    nJobs -= 1;
	    if (jobFull && DEBUG(JOB)) {
		printf("Job queue is no longer full.\n");
	    }
	    jobFull = FALSE;
	
	    if (job->flags & JOB_REMOTE) {
		Rmt_Done (job->rmtID);
	    } else {
		nLocal -= 1;
	    }
	}

	JobFinish (job, status);
	
	if (errors && !keepgoing && (aborting != ABORT_INTERRUPT)) {
	    /*
	     * If we found any errors in this batch of children and the -k flag
	     * wasn't given, we set the aborting flag so no more jobs get
	     * started.
	     */
	    aborting = ABORT_ERROR;
	}

	if ((aborting == ABORT_ERROR) && Job_Empty()) {
	    /*
	     * If we are aborting and the job table is now empty, we finish.
	     */
	    (void) unlink (tfile);
	    Finish (errors);
	}
    }
}

/*-
 *-----------------------------------------------------------------------
 * Job_CatchOutput --
 *	Catch the output from our children, if we're using
 *	pipes do so. Otherwise just block time until we get a
 *	signal (most likely a SIGCHLD) since there's no point in
 *	just spinning when there's nothing to do and the reaping
 *	of a child can wait for a while. 
 *
 * Results:
 *	None 
 *
 * Side Effects:
 *	Output is read from pipes if we're piping.
 * -----------------------------------------------------------------------
 */
void
Job_CatchOutput ()
{
    int           	  nfds;
    struct timeval	  timeout;
    fd_set           	  readfds;
    register LstNode	  ln;
    register Job   	  *job;
    int	    	  	  pnJobs;   	/* Previous nJobs */

#ifdef CUSTOMS
    pnJobs = nJobs;

    while (pnJobs == nJobs) {
	Rpc_Wait();
    }
#else
    if (usePipes) {
	readfds = outputs;
	timeout.tv_sec = SEL_SEC;
	timeout.tv_usec = SEL_USEC;

	if ((nfds = select (FD_SETSIZE, &readfds, (int *) 0, (int *) 0, &timeout)) < 0)
	{
	    return;
	} else {
	    if (Lst_Open (jobs) == FAILURE) {
		Punt ("Cannot open job table");
	    }
	    while (nfds && (ln = Lst_Next (jobs)) != NILLNODE) {
		job = (Job *) Lst_Datum (ln);
		if (FD_ISSET(job->inPipe, &readfds)) {
		    JobDoOutput (job, FALSE);
		    nfds -= 1;
		}
	    }
	    Lst_Close (jobs);
	}
    }
#endif /* CUSTOMS */
}

/*-
 *-----------------------------------------------------------------------
 * Job_Make --
 *	Start the creation of a target. Basically a front-end for
 *	JobStart used by the Make module.
 *
 * Results:
 *	None.
 *
 * Side Effects:
 *	Another job is started.
 *
 *-----------------------------------------------------------------------
 */
void
Job_Make (gn)
    GNode   *gn;
{
    (void)JobStart (gn, 0, (Job *)NULL);
}

/*-
 *-----------------------------------------------------------------------
 * Job_Init --
 *	Initialize the process module
 *
 * Results:
 *	none
 *
 * Side Effects:
 *	lists and counters are initialized
 *-----------------------------------------------------------------------
 */
void
Job_Init (maxproc, maxlocal)
    int           maxproc;  /* the greatest number of jobs which may be
			     * running at one time */
    int	    	  maxlocal; /* the greatest number of local jobs which may
			     * be running at once. */
{
    GNode         *begin;     /* node for commands to do at the very start */
#ifdef CUSTOMS
    struct timeval waitTime;	/* Interval at which to wake up to check for
				 * dead children */
    if (DEBUG(RMT)) {
	Rpc_Debug(1);
    }
#endif /* CUSTOMS */

#ifdef Sprite
    sprintf (tfile, "/tmp/make%05x", getpid());
#else
    sprintf (tfile, "/tmp/make%05d", getpid());
#endif /* Sprite */

    jobs =  	  Lst_Init (FALSE);
    stoppedJobs = Lst_Init(FALSE);
    maxJobs = 	  maxproc;
    maxLocal = 	  maxlocal;
    nJobs = 	  0;
    nLocal = 	  0;
    jobFull = 	  FALSE;

    aborting = 	  0;
    errors = 	  0;

    lastNode =	  NILGNODE;

    if (shellPath == (char *) NULL) {
	/*
	 * The user didn't specify a shell to use, so we are using the
	 * default one... Both the absolute path and the last component
	 * must be set. The last component is taken from the 'name' field
	 * of the default shell description pointed-to by commandShell.
	 * All default shells are located in DEFSHELLDIR.
	 */
	shellName = commandShell->name;
	shellPath = Str_Concat (DEFSHELLDIR, shellName, STR_ADDSLASH);
    }

    if (commandShell->exit == (char *)NULL) {
	commandShell->exit = "";
    }
    if (commandShell->echo == (char *)NULL) {
	commandShell->echo = "";
    }


#ifdef CUSTOMS
    /*
     * Set up an event to check if any children have died.
     */
    waitTime.tv_sec = SEL_SEC;
    waitTime.tv_usec = SEL_USEC;
    (void)Rpc_EventCreate(&waitTime, JobCheckChildren, (Rpc_Opaque)0);
#endif /* CUSTOMS */
    signal(SIGINT, JobInterrupt);

    begin = Targ_FindNode (".BEGIN", TARG_NOCREATE);

    if (begin != NILGNODE) {
	JobStart (begin, JOB_SPECIAL, (Job *)0);
	while (nJobs) {
	    Job_CatchOutput();
	    Job_CatchChildren (!usePipes);
	}
    }
    postCommands = Targ_FindNode (".END", TARG_CREATE);
}

/*-
 *-----------------------------------------------------------------------
 * Job_Full --
 *	See if the job table is full. It is considered full if it is OR
 *	if we are in the process of aborting OR if we have
 *	reached/exceeded our local quota. This prevents any more jobs
 *	from starting up.
 *
 * Results:
 *	TRUE if the job table is full, FALSE otherwise
 * Side Effects:
 *	None.
 *-----------------------------------------------------------------------
 */
Boolean
Job_Full ()
{
    return (aborting || jobFull);
}

/*-
 *-----------------------------------------------------------------------
 * Job_Empty --
 *	See if the job table is empty.  Because the local concurrency may
 *	be set to 0, it is possible for the job table to become empty,
 *	while the list of stoppedJobs remains non-empty. In such a case,
 *	we want to restart as many jobs as we can.
 *
 * Results:
 *	TRUE if it is. FALSE if it ain't.
 *
 * Side Effects:
 *	None.
 *
 * -----------------------------------------------------------------------
 */
Boolean
Job_Empty ()
{
    if (nJobs == 0) {
	if (!Lst_IsEmpty(stoppedJobs) && !aborting) {
	    /*
	     * The job table is obviously not full if it has no jobs in
	     * it...Try and restart the stopped jobs.
	     */
	    jobFull = FALSE;
	    while (!jobFull && !Lst_IsEmpty(stoppedJobs)) {
		JobRestart((Job *)Lst_DeQueue(stoppedJobs));
	    }
	    return(FALSE);
	} else {
	    return(TRUE);
	}
    } else {
	return(FALSE);
    }
}

/*-
 *-----------------------------------------------------------------------
 * JobMatchShell --
 *	Find a matching shell in 'shells' given its final component.
 *
 * Results:
 *	A pointer to the Shell structure.
 *
 * Side Effects:
 *	None.
 *
 *-----------------------------------------------------------------------
 */
static Shell *
JobMatchShell (name)
    char	  *name;      /* Final component of shell path */
{
    register Shell *sh;	      /* Pointer into shells table */
    Shell	   *match;    /* Longest-matching shell */
    register char *cp1,
		  *cp2;
    char	  *eoname;

    eoname = name + strlen (name);

    match = (Shell *) NULL;

    for (sh = shells; *sh->name != '\0'; sh++) {
	for (cp1 = eoname - strlen (sh->name), cp2 = sh->name;
	     *cp1 != '\0' && *cp1 == *cp2;
	     cp1++, cp2++) {
		 continue;
	}
	if (*cp1 != *cp2) {
	    continue;
	} else if (match == (Shell *) NULL ||
		   strlen (match->name) < strlen (sh->name)) {
		       match = sh;
	}
    }
    return (match == (Shell *) NULL ? sh : match);
}

/*-
 *-----------------------------------------------------------------------
 * Job_ParseShell --
 *	Parse a shell specification and set up commandShell, shellPath
 *	and shellName appropriately.
 *
 * Results:
 *	FAILURE if the specification was incorrect.
 *
 * Side Effects:
 *	commandShell points to a Shell structure (either predefined or
 *	created from the shell spec), shellPath is the full path of the
 *	shell described by commandShell, while shellName is just the
 *	final component of shellPath.
 *
 * Notes:
 *	A shell specification consists of a .SHELL target, with dependency
 *	operator, followed by a series of blank-separated words. Double
 *	quotes can be used to use blanks in words. A backslash escapes
 *	anything (most notably a double-quote and a space) and
 *	provides the functionality it does in C. Each word consists of
 *	keyword and value separated by an equal sign. There should be no
 *	unnecessary spaces in the word. The keywords are as follows:
 *	    name  	    Name of shell.
 *	    path  	    Location of shell. Overrides "name" if given
 *	    quiet 	    Command to turn off echoing.
 *	    echo  	    Command to turn echoing on
 *	    filter	    Result of turning off echoing that shouldn't be
 *	    	  	    printed.
 *	    echoFlag	    Flag to turn echoing on at the start
 *	    errFlag	    Flag to turn error checking on at the start
 *	    hasErrCtl	    True if shell has error checking control
 *	    check 	    Command to turn on error checking if hasErrCtl
 *	    	  	    is TRUE or template of command to echo a command
 *	    	  	    for which error checking is off if hasErrCtl is
 *	    	  	    FALSE.
 *	    ignore	    Command to turn off error checking if hasErrCtl
 *	    	  	    is TRUE or template of command to execute a
 *	    	  	    command so as to ignore any errors it returns if
 *	    	  	    hasErrCtl is FALSE.
 *
 *-----------------------------------------------------------------------
 */
ReturnStatus
Job_ParseShell (line)
    char	  *line;  /* The shell spec */
{
    char    	  **words;
    int	    	  wordCount;
    register char **argv;
    register int  argc;
    char    	  *path;
    Shell   	  newShell;
    Boolean 	  fullSpec = FALSE;

    while (isspace (*line)) {
	line++;
    }
    words = Str_BreakString (line, " \t", "\n", &wordCount);

    bzero ((Address)&newShell, sizeof(newShell));
    
    /*
     * Parse the specification by keyword
     */
    for (path = (char *)NULL, argc = wordCount - 1, argv = words + 1;
	 argc != 0;
	 argc--, argv++) {
	     if (strncmp (*argv, "path=", 5) == 0) {
		 path = &argv[0][5];
	     } else if (strncmp (*argv, "name=", 5) == 0) {
		 newShell.name = &argv[0][5];
	     } else {
		 if (strncmp (*argv, "quiet=", 6) == 0) {
		     newShell.echoOff = &argv[0][6];
		 } else if (strncmp (*argv, "echo=", 5) == 0) {
		     newShell.echoOn = &argv[0][5];
		 } else if (strncmp (*argv, "filter=", 7) == 0) {
		     newShell.noPrint = &argv[0][7];
		     newShell.noPLen = strlen(newShell.noPrint);
		 } else if (strncmp (*argv, "echoFlag=", 9) == 0) {
		     newShell.echo = &argv[0][9];
		 } else if (strncmp (*argv, "errFlag=", 8) == 0) {
		     newShell.exit = &argv[0][8];
		 } else if (strncmp (*argv, "hasErrCtl=", 10) == 0) {
		     char c = argv[0][10];
		     newShell.hasErrCtl = !((c != 'Y') && (c != 'y') &&
					    (c != 'T') && (c != 't'));
		 } else if (strncmp (*argv, "check=", 6) == 0) {
		     newShell.errCheck = &argv[0][6];
		 } else if (strncmp (*argv, "ignore=", 7) == 0) {
		     newShell.ignErr = &argv[0][7];
		 } else {
		     Parse_Error (PARSE_FATAL, "Unknown keyword \"%s\"",
				  *argv);
		     Str_FreeVec (wordCount, words);
		     return (FAILURE);
		 }
		 fullSpec = TRUE;
	     }
    }

    if (path == (char *)NULL) {
	/*
	 * If no path was given, the user wants one of the pre-defined shells,
	 * yes? So we find the one s/he wants with the help of JobMatchShell
	 * and set things up the right way. shellPath will be set up by
	 * Job_Init.
	 */
	if (newShell.name == (char *)NULL) {
	    Parse_Error (PARSE_FATAL, "Neither path nor name specified");
	    Str_FreeVec (wordCount, words);
	    return (FAILURE);
	} else {
	    commandShell = JobMatchShell (newShell.name);
	    shellName = newShell.name;
	}
    } else {
	/*
	 * The user provided a path. If s/he gave nothing else (fullSpec is
	 * FALSE), try and find a matching shell in the ones we know of.
	 * Else we just take the specification at its word and copy it
	 * to a new location. In either case, we need to record the
	 * path the user gave for the shell.
	 */
	shellPath = path;
	path = rindex (path, '/');
	if (path == (char *)NULL) {
	    path = shellPath;
	} else {
	    path += 1;
	}
	if (newShell.name != (char *)NULL) {
	    shellName = newShell.name;
	} else {
	    shellName = path;
	}
	if (!fullSpec) {
	    commandShell = JobMatchShell (shellName);
	} else {
	    commandShell = (Shell *) malloc(sizeof(Shell));
	    *commandShell = newShell;
	}
    }

    if (commandShell->echoOn && commandShell->echoOff) {
	commandShell->hasEchoCtl = TRUE;
    }
    
    if (!commandShell->hasErrCtl) {
	if (commandShell->errCheck == (char *)NULL) {
	    commandShell->errCheck = "";
	}
	if (commandShell->ignErr == (char *)NULL) {
	    commandShell->ignErr = "%s\n";
	}
    }
    
    /*
     * Do not free up the words themselves, since they might be in use by the
     * shell specification...
     */
    free (words);
    return SUCCESS;
}

/*-
 *-----------------------------------------------------------------------
 * JobInterrupt --
 *	Handle the receipt of an interrupt.
 *
 * Results:
 *	None
 *
 * Side Effects:
 *	All children are killed. Another job will be started if the
 *	.INTERRUPT target was given.
 *-----------------------------------------------------------------------
 */
static int
JobInterrupt ()
{
    LstNode 	  ln;		/* element in job table */
    Job           *job;	    	/* job descriptor in that element */
    GNode         *interrupt;	/* the node describing the .INTERRUPT target */
    
    aborting = ABORT_INTERRUPT;

    (void)Lst_Open (jobs);
    while ((ln = Lst_Next (jobs)) != NILLNODE) {
	job = (Job *) Lst_Datum (ln);

	if (!Targ_Precious (job->node)) {
	    char  	*file = (job->node->path == (char *)NULL ?
				 job->node->name :
				 job->node->path);
	    if (unlink (file) == 0) {
		Error ("*** %s removed", file);
	    }
	}
#ifdef CUSTOMS
	if (job->flags & JOB_REMOTE) {
	    /*
	     * If job is remote, we need to send the signal to the remote
	     * agent to kill the job for us. Form a packet and the server's
	     * address from the export permit we kept with the job and
	     * call the agent's CUSTOMS_KILL procedure. If we get an error from
	     * it, we assume the job is done and finish it out with an error
	     * code.
	     */
	    Kill_Data	packet;
	    struct sockaddr_in	server;

	    packet.id = job->permit.id;
	    packet.signo = SIGINT;
	    server.sin_family = AF_INET;
	    server.sin_addr = job->permit.addr;
	    server.sin_port = customs_AgentAddr.sin_port;
	    
	    if (Rpc_Call(customs_Socket, &server, (Rpc_Proc)CUSTOMS_KILL,
			 sizeof(packet), (Rpc_Opaque)&packet,
			 0, (Rpc_Opaque)0,
			 CUSTOMS_NRETRY, &customs_RetryTimeOut) != RPC_SUCCESS)
	    {
		/*
		 * If couldn't kill the thing, finish it out now with an
		 * error code.
		 */
		union wait status;

		status.w_status = 0;
		status.w_retcode = 1;
		JobFinish(job, status);
	    } else {
		/*
		 * We killed the beast, so finish it off now (so we don't have
		 * to call Rpc_Wait down below)
		 */
		union wait status;

		status.w_status = 0;
		status.w_termsig = SIGINT;

		JobFinish(job, status);
	    }
	} else
#endif /* CUSTOMS */
	if (job->pid) {
#ifdef USE_PGRP
# ifdef SYSV
	    kill(-job->pid, SIGINT);
# else
	    killpg(job->pid, SIGINT);
# endif
#else
	    kill(job->pid, SIGINT);
#endif /* USE_PGRP */
	}
    }
    Lst_Close (jobs);

    if (!touchFlag) {
	interrupt = Targ_FindNode (".INTERRUPT", TARG_NOCREATE);
	if (interrupt != NILGNODE) {
	    ignoreErrors = FALSE;

	    JobStart (interrupt, JOB_IGNDOTS, (Job *)0);
	    while (nJobs) {
		Job_CatchOutput();
		Job_CatchChildren (!usePipes);
	    }
	}
    }
    (void) unlink (tfile);
    exit (0);
}

/*
 *-----------------------------------------------------------------------
 * Job_End --
 *	Do final processing such as the running of the commands
 *	attached to the .END target. 
 *
 * Results:
 *	Number of errors reported.
 *
 * Side Effects:
 *	The process' temporary file (tfile) is removed if it still
 *	existed.
 *-----------------------------------------------------------------------
 */
int
Job_End ()
{
    if (postCommands != NILGNODE && !Lst_IsEmpty (postCommands->commands)) {
	if (errors) {
	    Error ("Errors reported so .END ignored");
	} else {
	    JobStart (postCommands, JOB_SPECIAL | JOB_IGNDOTS,
		       (Job *)0);

	    while (nJobs) {
		Job_CatchOutput();
		Job_CatchChildren (!usePipes);
	    }
	}
    }
    (void) unlink (tfile);
    return(errors);
}

/*-
 *-----------------------------------------------------------------------
 * Job_Wait --
 *	Waits for all running jobs to finish and returns. Sets 'aborting'
 *	to ABORT_WAIT to prevent other jobs from starting.
 *
 * Results:
 *	None.
 *
 * Side Effects:
 *	Currently running jobs finish.
 *
 *-----------------------------------------------------------------------
 */
void
Job_Wait()
{
    aborting = ABORT_WAIT;
    while (nJobs != 0) {
	Job_CatchOutput();
	Job_CatchChildren(!usePipes);
    }
    aborting = 0;
}

/*-
 *-----------------------------------------------------------------------
 * Job_AbortAll --
 *	Abort all currently running jobs without handling output or anything.
 *	This function is to be called only in the event of a major
 *	error. Most definitely NOT to be called from JobInterrupt.
 *
 * Results:
 *	None
 *
 * Side Effects:
 *	All children are killed, not just the firstborn
 *-----------------------------------------------------------------------
 */
void
Job_AbortAll ()
{
    LstNode           	ln;		/* element in job table */
    Job            	*job;	/* the job descriptor in that element */
    int     	  	foo;
    
    aborting = ABORT_ERROR;
    
    if (nJobs) {

	(void)Lst_Open (jobs);
	while ((ln = Lst_Next (jobs)) != NILLNODE) {
	    job = (Job *) Lst_Datum (ln);

#ifdef CUSTOMS
	    if (job->flags & JOB_REMOTE) {
		/*
		 * If job is remote, we need to send the signal to the remote
		 * agent to kill the job for us. Form a packet and the server's
		 * address from the export permit we kept with the job and
		 * call the agent's CUSTOMS_KILL procedure. If we get an error
		 * from it, we assume the job is done and finish it out with an
		 * error code.
		 */
		Kill_Data	packet;
		struct sockaddr_in	server;
		
		packet.id = job->permit.id;
		packet.signo = SIGINT;
		server.sin_family = AF_INET;
		server.sin_addr = job->permit.addr;
		server.sin_port = customs_AgentAddr.sin_port;
		
		if (Rpc_Call(customs_Socket, &server, (Rpc_Proc)CUSTOMS_KILL,
			     sizeof(packet), (Rpc_Opaque)&packet,
			     0, (Rpc_Opaque)0,
			     CUSTOMS_NRETRY,
			     &customs_RetryTimeOut) == RPC_SUCCESS)
		{
		    packet.signo = SIGKILL;
		    (void)Rpc_Call(customs_Socket, &server,
				   (Rpc_Proc)CUSTOMS_KILL,
				   sizeof(packet), (Rpc_Opaque)&packet,
				   0, (Rpc_Opaque)0,
				   CUSTOMS_NRETRY,
				   &customs_RetryTimeOut);
		}
	    } else {
#endif /* CUSTOMS */
		
	    /*
	     * kill the child process with increasingly drastic signals to make
	     * darn sure it's dead. 
	     */
#ifdef USE_PGRP
# ifdef SYSV
		kill(-job->pid, SIGINT);
		kill(-job->pid, SIGKILL);
# else
		killpg(job->pid, SIGINT);
		killpg(job->pid, SIGKILL);
# endif
#else
		kill(job->pid, SIGINT);
		kill(job->pid, SIGKILL);
#endif /* USE_PGRP */
#ifdef CUSTOMS
	    }
#endif /* CUSTOMS */
	}
    }
    /*
     * Catch as many children as want to report in at first, then give up
     */
    while (wait3(&foo, WNOHANG, (struct rusage *)0) > 0) {
	;
    }
    (void) unlink (tfile);
}
